import requests
import base64
import json
import logging

class RUBYXApi:
    RUBYX_API = None
    auth = None
    doctypes = ['ip', 'domain', 'program', 'agent', 'url', 'service', 'config', 'alert']
    sc = None
    discord_webhook = None
    do_debug = False
    requests_session = None
    
    def __init__(self, couchdb_url, user, pwd, discord_webhook = None, ignore_ssl_errors = False, debug = False):
        auth = user+':'+pwd
        self.auth = 'Basic '+base64.b64encode(auth.encode('utf-8')).decode('utf-8')
        self.requests_session = requests.Session()
        self.do_debug = debug
        if debug:
            logging.basicConfig()
            logging.getLogger().setLevel(logging.DEBUG)
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True
        if discord_webhook:
            self.discord_webhook = discord_webhook
        if ignore_ssl_errors:
            requests.packages.urllib3.disable_warnings()
            self.requests_session.verify = False
        self.RUBYX_API = couchdb_url
    
    '''
    Create a new program.
    '''
    def create_new_program(self, program_name=None, disabled=False, passive_only=False, tags=[]):
        if not program_name:
            return
        else:
            program = {"type": "program", "disabled": disabled, "passive_only": passive_only, "inscope": [], "outscope": []}
            if tags:
                program['tags'] = self.process_tags(tags)
            r = self.requests_session.put(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe=''), json.dumps(program), headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
    
    '''
    Get a list of all domains, filtered by program name if provided.
    
    @todo: generalize this and get_ips by general function get_document_by_program_name with additional paramter doctype
    '''
    def get_domains_by_program_name(self, program_name=None, show_disabled=False):
        if program_name:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/domains?reduce=false&key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
            return [r['value'] for r in r.json()['rows']]
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/domains?reduce=false', headers={"Authorization": self.auth})
            # by default, filter out all disabled programs - sadly this needs to happen client side;
            # because there is no relational link between a domain and its program to look up the status of 
            # the program when indexing.
            # Actually, this can probably be improved with custom queries:
            # `https://docs.couchdb.org/en/stable/api/database/find.html`
            if show_disabled:
                # no filter, so return everything
                return [r['value'] for r in r.json()['rows']]
            else:
                active_programs = self.get_programs(show_disabled=show_disabled)
                return [r['value'] for r in r.json()['rows'] if r['key'] in active_programs]
            
    
    '''
    Get a list of all urls, filtered by program or hostname if provided.
    '''
    def get_urls_by_hostname(self, hostname=None, with_query=False, root_only=False):
        if hostname:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/urls_by_hostname?reduce=false&key="'+hostname+'"', headers={"Authorization": self.auth})
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/urls_by_hostname?reduce=false', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        # print all url, status, content_length if status and content length are set    
        return self.process_urls(r.json()['rows'], with_query, root_only)
    
    def get_urls_by_program(self, program=None, with_query=False, root_only=False, show_disabled=False):
        if program:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/urls_by_program?reduce=false&key="'+requests.utils.quote(program, safe='')+'"', headers={"Authorization": self.auth})
            return self.process_urls(r.json()['rows'], with_query, root_only)
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/urls_by_program?reduce=false', headers={"Authorization": self.auth})
            if show_disabled:
                # no filter, so return everything
                return self.process_urls(r.json()['rows'], with_query)
            else:
                active_programs = self.get_programs(show_disabled=show_disabled)
                return self.process_urls([x for x in r.json()['rows'] if x['key'] in active_programs], with_query, root_only)        
        
    def process_urls(self, urls, with_query=False, root_only=False):
        if not with_query:
            if not root_only:
            # print all url, status, content_length if status and content length are set
                return [" ".join([str(x) for x in r['value'][:3] if x]) for r in urls]
            else:
            # print only roots or urls:
                return list(set(['/'.join(r['value'][0].split('/',3)[:3]) for r in urls]))
        else:
            # get a list of URLs without any queries
            no_query = [" ".join([str(x) for x in r['value'][:3] if x]) for r in urls if len(r['value'][3]) == 0]
            # expand the URLs that do have queries
            expanded = [ [url['value'][0]+'?'+q] + url['value'][1:3] for url in urls for q in url['value'][3] if q]
            return [" ".join([str(x) for x in r[:3] if x]) for r in expanded] + no_query
    
    '''
    Get a list of all services, filtered by program name if provided.
    '''
    def get_services_by_program_name(self, program_name=None, show_disabled=False):
        if program_name:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/services?reduce=false&key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
            return [r['value'] for r in r.json()['rows']]
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/services?reduce=false', headers={"Authorization": self.auth})
            if show_disabled:
                # no filter, so return everything
                return [r['value'] for r in r.json()['rows']]
            else:
                active_programs = self.get_programs(show_disabled=show_disabled)
                return [r['value'] for r in r.json()['rows'] if r['key'] in active_programs]
    
    '''
    Get all documents of a certain type
    '''
    def get_documents_of_type(self, doctype, program_name = None):
        if doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        if doctype == 'agent':
            return self.get_agents()
        if program_name:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/'+doctype+'s?reduce=false&key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/'+doctype+'s?reduce=false', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        return [r['value'] for r in r.json()['rows']]
    
    '''
    @todo put restrictions on 'view' parameter, should only allow lowercase and _
    '''
    def get_documents_view(self, program_name, doctype, view, show_disabled=False):
        if doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        if program_name:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/'+doctype+'s_'+view+'?reduce=false&key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
            return [r['value'] for r in r.json()['rows']]
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/'+doctype+'s_'+view+'?reduce=false', headers={"Authorization": self.auth})
            if show_disabled:
                # no filter, so return everything
                return [r['value'] for r in r.json()['rows']]
            else:
                active_programs = self.get_programs(show_disabled=show_disabled)
                return [r['value'] for r in r.json()['rows'] if r['key'] in active_programs]
    
    '''
    Get a list of all ips, filtered by program name if provided.
    '''
    def get_ips_by_program_name(self, program_name=None, filter_cdn=False, show_disabled=False):
        
        cdn_filter = '_no_cdn' if filter_cdn else ''
        
        if program_name:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/ips'+cdn_filter+'?reduce=false&key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
            return [r['value'] for r in r.json()['rows']]
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/ips'+cdn_filter+'?reduce=false', headers={"Authorization": self.auth})
            # by default, filter out all disabled programs
            if show_disabled:
                return [r['value'] for r in r.json()['rows']]
            else:
                active_programs = self.get_programs(show_disabled=show_disabled)
                return [r['value'] for r in r.json()['rows'] if r['key'] in active_programs]
    
    '''
    Get a list of all programs.
    '''
    def get_programs(self, show_disabled=False, show_empty_scope=False):
        if show_disabled:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/programs?reduce=false', headers={"Authorization": self.auth})
        else:
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/programs?reduce=false&key=false', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        
        try:
            if show_empty_scope:
                return [r['id'] for r in r.json()['rows']]
            else:
                return [r['id'] for r in r.json()['rows'] if r['value'] > 0]
        except:
            print('[WARNING] Your RUBYX server views are deprecated, it is strongly recommended to upgrade. Run `rubyx server upgrade` to upgrade.')
            return [r['value'] for r in r.json()['rows']]
    
    def get_program_scope(self, program_name):
        if program_name == '@INFER':
            return None, [], []
        self.debug('getting program scope')
        r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe=''), headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        return r.json(), r.json()['inscope'], r.json()['outscope']
    
    def update_program_scope(self, program_name, inscope, outscope, program=None):
        self.debug('updating program scope')
        if not program:
            r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe=''), headers={"Authorization": self.auth})
            if 'error' in r.json():
                raise Exception('RUBYX server error: '+r.json()['error'])
            program = r.json()
        # only update if different!
        if program['inscope'] != inscope or program['outscope'] != outscope:
            program['inscope'] = inscope
            program['outscope'] = outscope
            r = self.requests_session.put(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe='')+'?rev='+program['_rev'], json.dumps(program), headers={"Authorization": self.auth})
            if 'error' in r.json():
                raise Exception('RUBYX server error: '+r.json()['error'])
            
    def get_program_blacklist(self, program_name=None, doc=None):
        self.debug('getting program blacklist')
        if program_name == '@INFER':
            return []
        if not doc:
            r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe=''), headers={"Authorization": self.auth})
            doc = r.json()
        if 'error' in doc:
            raise Exception('RUBYX server error: '+r.json()['error'])
        if 'blacklist' not in doc:
            return []
        return doc['blacklist']
    
    def update_program_blacklist(self, program_name, blacklist):
        self.debug('updating program blacklist')
        r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe=''), headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        program = r.json()
        program['blacklist'] = blacklist
        r = self.requests_session.put(self.RUBYX_API+'/'+requests.utils.quote(program_name, safe='')+'?rev='+program['_rev'], json.dumps(program), headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        
    '''
    Return all document identifiers of documents belonging to a program.
    '''
    def get_all_program_documents(self, program_name):
        r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/program_all_documents?key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        return [ (r['id'], r['value']) for r in r.json()['rows'] ]
            
    '''
    Get a list of all agents.
    '''
    def get_agents(self):
        self.debug('getting agents')
        r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/agents', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        return r.json()['rows']
    
    '''
    Register a new agent.
    '''
    def register_agent(self, name):
        self.debug('registering agent')
        r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/agents', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        return [r['key'] for r in r.json()['rows']]
    
    '''
    Add a list of documents to a program in bulk.
    '''
    def add_documents(self, doctype, identifiers, program_name, source=None, tags=[]):
        self.debug('adding documents in bulk')
        if doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        # Create documents in bulk
        bulk = {"docs": []}
        doc = {}
        
        relname = ''
        if doctype == 'ip':
            relname = 'domains'
        elif doctype == 'domain':
            relname = 'ips'
        
        if program_name == '@INFER':
            # When related identifiers are supplied, e.g. `rubyx ip add 1.1.1.1:example.com`
            # we can infer the program by fetching the program for `example.com`
            
            program_candidates = {}
            
            if relname:
                # get all candidate program names by identifier in rel:
                r = self.requests_session.post(self.RUBYX_API+'/_all_docs?include_docs=true', json.dumps({'keys': [x for docid in identifiers.keys() for x in identifiers[docid]]}), headers={"Authorization": self.auth, 'Content-Type': 'application/json'})
                program_candidates = {a['key']: a['doc']['program'] for a in r.json()['rows'] if 'doc' in a and a['doc']}
            
            # get all scope definitions of programs to dynamically match domains
            r = self.requests_session.post(self.RUBYX_API+'/_all_docs?include_docs=true', json.dumps({'keys': [x for x in self.get_programs()]}), headers={"Authorization": self.auth, 'Content-Type': 'application/json'})
            program_scopes = {r['id']: {'in': r['doc']['inscope'], 'out':r['doc']['outscope']}  for r in r.json()['rows'] if 'doc' in r and r['doc']}
        
        for docid in identifiers.keys():
            # extract the appropriate program name from the list of candidates
            infered_program_name = program_name
            if program_name == '@INFER':
                infered_program_name = [program_candidates[d] for d in program_candidates.keys() if d in identifiers[docid]]
                if len(infered_program_name) > 0:
                    infered_program_name = infered_program_name[0]
                else:
                    # no luck with the linked documents, but if we're adding a URL or domain,
                    # we may match against known scopes
                    if doctype == 'domain':
                        from . import RubyxClient
                        infered_program_name = [p for p in program_scopes.keys() if RubyxClient.matches_scope(docid, program_scopes[p]['in']) and not RubyxClient.matches_scope(docid, program_scopes[p]['out'])]
                        if len(infered_program_name) > 0:
                            infered_program_name = infered_program_name[0]
                    elif doctype == 'url':
                        from . import RubyxClient
                        infered_program_name = [p for p in program_scopes.keys() if RubyxClient.matches_scope(identifiers[docid]['hostname'], program_scopes[p]['in']) and not RubyxClient.matches_scope(identifiers[docid]['hostname'], program_scopes[p]['out'])]
                        if len(infered_program_name) > 0:
                            infered_program_name = infered_program_name[0]

            if not infered_program_name or infered_program_name == '@INFER':
                # was unable to infer, skipping document
                continue
                    
            doc = {
                '_id': docid,
                'program': infered_program_name,
                'type': doctype
            }
            if relname:
                if len(identifiers[docid])>0:
                    doc[relname] = list(filter(None,identifiers[docid]))  # filter out the empty values
                else:
                    doc[relname] = []
            
            else:
                # in order to support urls and general flexibility,
                # this needs to support arbitrary keys to be added to the JSON document
                for key in identifiers[docid]:
                    if key not in ['_rev', '_id']:
                        doc[key] = identifiers[docid][key]
                
            if source:
                doc['source'] = source
                
            if tags:
                doc['tags'] = self.process_tags(tags)
                
            bulk['docs'].append(doc)
        
        r = self.requests_session.post(
            self.RUBYX_API+'/_bulk_docs', json.dumps(bulk),
            headers={"Authorization": self.auth, "Content-Type": "application/json"}
        )
        
        if r.status_code == 504:
            raise Exception('Server timed out - please check if the operation succeeded or repeat')

        if 'error' in r.json():
            raise Exception('Unexpected RUBYX response: '+r.json()['error'])
        
        # return (success,failed) with identifiers of new docs and docs that failed due to conflict
        return (
            [doc['id'] for doc in r.json() if 'ok' in doc and doc['ok']],
            [doc['id'] for doc in r.json() if 'error' in doc and doc['error'] == 'conflict']
        )
        
        
    
    '''
    Get the identifier of a document based on a set of properties defined in propmap
    '''
    def get_document_id_by_properties(self, doctype, propmap):
        if doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        docs = self.get_documents_of_type(doctype)
        for doc in docs:
            matches = True
            for prop in propmap.keys():
                if prop not in doc or propmap[prop] != doc[prop]:
                    matches = False
            if matches:
                return doc['id']
                
    '''
    Remove a document from the database, based on a document property.
    '''
    def remove_document(self, doctype, document):
        if doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
            
        if type(document) is dict:
            document = self.get_document_id_by_properties(doctype, document)
            # print(document)
            
        # Need to encode the document so it can handle CIDR ranges including /
        r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(document, safe=''), headers={"Authorization": self.auth})
        if 'error' in r.json() and r.json()['error'] != 'not_found':
            raise Exception('RUBYX server error: '+r.json()['error'])
        elif 'error' in r.json() and r.json()['error'] == 'not_found':
            return
        if 'type' in r.json() and not r.json()['type'] == doctype:
            raise Exception('The specified document (type: '+r.json()['type']+') is not of the requested type '+doctype)
        if '_rev' in r.json():
            r = self.requests_session.delete(self.RUBYX_API+'/'+requests.utils.quote(document, safe='')+'?rev='+r.json()['_rev'], headers={"Authorization": self.auth})
            if 'error' in r.json() and r.json()['error'] != 'not_found':
                raise Exception('RUBYX server error: '+r.json()['error'])

    '''
    This is a dangerous function - will always require user confirmation
    to remove the list of supplied documents
    '''
    def remove_documents(self, docids):
        self.debug("Removing documents in bulk")

        # first, get the latest revisions of each of the listed documents
        r = self.requests_session.post(self.RUBYX_API+'/_all_docs?include_docs=true', json.dumps({'keys': [x for x in docids]}), headers={"Authorization": self.auth, 'Content-Type': 'application/json'})
        to_be_removed = [{'_id': x['key'], '_rev': x['value']['rev'], '_deleted': True} for x in r.json()['rows'] if 'doc' in x and x['doc']]

        # now bulk remove all documents!
        if len(to_be_removed) > 0:
            r = self.requests_session.post(
                self.RUBYX_API+'/_bulk_docs',
                json.dumps({'docs': to_be_removed}),
                headers={"Authorization": self.auth, 'Content-Type': 'application/json'}
            )
            if 'error' in r.json():
                raise Exception('RUBYX server error: '+r.json()['error'])

        return [x['_id'] for x in to_be_removed]
                
    '''
    Return a raw version of a document by id
    '''
    def get_document(self, docid):
        # escape leading underscores
        if docid.startswith('_'):
            docid = '.'+docid
        r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(docid, safe=''), headers={"Authorization": self.auth})
        if 'error' in r.json() and r.json()['error'] == 'not_found':
            return None
        return r.text
    
    '''
    Return a JSON of many documents by id
    '''
    def get_documents(self, docids):
        
        r = self.requests_session.post(self.RUBYX_API+'/_all_docs?include_docs=true', json.dumps({'keys': [x if not x.startswith('_.') else '.'+x for x in docids]}), headers={"Authorization": self.auth, 'Content-Type': 'application/json'})
        
        if 'error' in r.json() and r.json()['error'] == 'not_found':
            return None
        return json.dumps([ x['doc'] for x in r.json()['rows'] if not 'error' in x ])
        
    '''
    Update a document, based on the document identifier.
    updates contains a map with keyname => [] that will
    be appended to the document.
    '''
    def update_document(self, doctype, document, updates, create_if_not_exists = False):
        if doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        r = self.requests_session.get(self.RUBYX_API+'/'+requests.utils.quote(document, safe=''), headers={"Authorization": self.auth})
        if 'error' in r.json() and r.json()['error'] != 'not_found':
            raise Exception('RUBYX server error: '+r.json()['error'])
        elif 'error' in r.json() and r.json()['error'] == 'not_found' and not create_if_not_exists:
            return
        if 'type' in r.json() and not r.json()['type'] == doctype:
            raise Exception('The specified document (type: '+r.json()['type']+') is not of the requested type '+doctype)
        if '_rev' in r.json():
            
            document_changed = False
            
            original_document = r.json()
            for prop in updates.keys():
                if prop not in original_document:
                    print('[Warning] The update specifies a key that does not exist: '+prop)
                    if type(updates[prop]) is list:
                        original_document[prop] = []  # initiate the property list if it doesn't exist
                    if type(updates[prop]) is dict:
                        original_document[prop] = {}  # initiate the property dict if it doesn't exist
                    else:
                        original_document[prop] = None
                if not type(original_document[prop]) is type(updates[prop]):
                    print('The updated key does not have the right type. Cannot update data of '+prop)
                    continue
                # if it's a list, append
                if type(original_document[prop]) is list:
                    for val in updates[prop]:         
                        if val not in original_document[prop]:
                            document_changed = True
                            original_document[prop].append(val)
                    # filter out empty values
                    original_document[prop] = list(filter(None, original_document[prop]))
                # if it's a dict:
                if type(original_document[prop]) is dict:
                    for val in updates[prop]:
                        if val not in original_document[prop]:
                            document_changed = True
                            original_document[prop][val] = updates[prop][val]
                        elif original_document[prop][val] != updates[prop][val]:
                            document_changed = True
                            original_document[prop][val] = updates[prop][val]
                # if it's not a list or a dict, replace
                elif not isinstance(original_document[prop], (list, dict)):
                    if not original_document[prop] == updates[prop]:
                        document_changed = True
                        original_document[prop] = updates[prop]
                        
            if document_changed:
                r = self.requests_session.put(
                    self.RUBYX_API+'/'+requests.utils.quote(document, safe='')+'?rev='+r.json()['_rev'],
                    json.dumps(original_document),
                    headers={"Authorization": self.auth, "Content-Type": "application/json"}
                )
                
                if 'error' in r.json():
                    raise Exception('RUBYX server error: '+r.json()['error'])
                
                return document
        elif create_if_not_exists and r.json()['error'] == 'not_found':
            # create the document for the first time
            r = self.requests_session.put(
                self.RUBYX_API+'/'+requests.utils.quote(document, safe=''),
                json.dumps(updates),
                headers={"Authorization": self.auth, "Content-Type": "application/json"}
            )
            
            if 'error' in r.json():
                raise Exception('RUBYX server error: '+r.json()['error'])
            
            return document
        
    '''
    Update documents in bulk
    The provided batch_updates need to be a dict of updates mapping the id to the dict of updates
    '''
    def update_documents(self, doctype, batch_updates, append_tags=False):
        self.debug('updating documents in bulk')
        # first, get the latest revisions of each of the listed documents
        r = self.requests_session.post(self.RUBYX_API+'/_all_docs?include_docs=true', json.dumps({'keys': [x for x in batch_updates.keys()]}), headers={"Authorization": self.auth, 'Content-Type': 'application/json'})
        
        current = {a['key']: a['doc'] for a in r.json()['rows'] if 'doc' in a and a['doc']}
        updates = {x['key']: batch_updates[x['key']].update({'_id': x['key'], '_rev':x['value']['rev']}) or batch_updates[x['key']] for x in r.json()['rows'] if 'doc' in x and x['doc']}
        
        to_be_updated = []
        
        for x in current.keys():
            # we need to ensure the source property is preserved when it's not explicitly set
            if not 'source' in updates[x] and 'source' in current[x]:
                updates[x]['source'] = current[x]['source']
            if not self.docs_are_equal(current[x], updates[x]):
                to_be_updated.append(updates[x])
                
        # TODO:
        # check current doctype against document.type
        for updates in to_be_updated:
            original_document = current[updates['_id']]
            
            if not original_document['type'] == doctype:
                print('[Error] The document type of document '+updates['_id']+' is not '+doctype)
                updates = {}
                continue
            
            # first ensure that all original properties are copied, so an update does not result in lost data!
            for prop in original_document.keys():
                if not prop in updates:
                    updates[prop] = original_document[prop]
            
            # now start changing what needs to be changed
            for prop in updates.keys():
                if prop not in original_document:
                    pass # this is fine, will happen e.g. when adding a status code or content_length for the first time
                else:
                    if not type(original_document[prop]) is type(updates[prop]):
                        print('[Error] The updated property '+prop+' doesn\'t have the right type. Cannot update '+updates['_id'])
                        updates = {}
                        continue
                    # if it's a list, make sure the update values are appended to the existing fields
                    if type(original_document[prop]) is list:
                        new_list = original_document[prop]
                        for val in updates[prop]:
                            if val not in new_list:
                                new_list.append(val)

                        # filter out empty values
                        updates[prop] = list(filter(None, new_list))
                        
                    # if it's a dict, make sure the update values are added to the existing map
                    # for example, this is the case when updating tags of a document
                    elif type(original_document[prop]) is dict:
                        new_dict = original_document[prop]
                        for key in updates[prop]:
                            if prop == 'tags' and append_tags:
                                # append the new value or list to the existing list
                                if not key in new_dict:
                                    new_dict[key] = updates[prop][key]
                                elif type(new_dict[key]) is list:
                                    new_dict[key].extend(updates[prop][key] if type(updates[prop][key]) is list else [updates[prop][key]])
                                else: # already existed with a single value
                                    updates[prop][key] = updates[prop][key] if type(updates[prop][key]) is list else [updates[prop][key]]
                                    updates[prop][key].extend([new_dict[key]])
                                    new_dict[key] = updates[prop][key]
                                # remove duplicates
                                if type(new_dict[key]) is list:
                                    new_dict[key] = list(set(new_dict[key]))
                            else:
                                # overwrite whatever values we have with the new value or list
                                new_dict[key] = updates[prop][key]
                        
                        # filter out empty values
                        updates[prop] = {x: new_dict[x] for x in new_dict if new_dict[x] and len(new_dict[x]) > 0}
                                
        
        # now bulk update all documents with changes!
        if len(to_be_updated) > 0:
            r = self.requests_session.post(
                self.RUBYX_API+'/_bulk_docs',
                json.dumps({'docs': to_be_updated}),
                headers={"Authorization": self.auth, 'Content-Type': 'application/json'}
            )
            if 'error' in r.json():
                raise Exception('RUBYX server error: '+r.json()['error'])

            return [x['_id'] for x in to_be_updated]
        
    '''
    return true if docs are identical, false if there are differences
    '''
    def docs_are_equal(self, current, updated, ignore = ['_id', '_rev', 'program', 'type'], check_lists = True):
        current = {k: current[k] for k in current if k not in ignore}
        updated = {k: updated[k] for k in updated if k not in ignore}
        
        if current == updated:
            return True
        
        # if documents are not equal and check_lists is set, we will also consider
        # documents equal if the elements of the update list are already in the current list
        if check_lists:
            current_nolists = {k: current[k] for k in current if k not in ignore and not type(current[k]) is list}
            updated_nolists = {k: updated[k] for k in updated if k not in ignore and not type(updated[k]) is list}
            
            # if even without the lists they are still different, don't bother to check the lists
            if not current_nolists == updated_nolists:
                return False
            else:
                update_lists = {k: updated[k] for k in updated if k not in ignore and type(updated[k]) is list}
                for lst in update_lists:
                    for l in update_lists[lst]:
                        if not l in current[lst]:
                            return False
                
                return True
        
        return False
        
    '''
    Inspired by https://github.com/dpavlin/angular-mojolicious/edit/master/couchdb-changes.pl
    
    Will handle changes in batches of 10 seconds at a time.
    '''
    def listen_for_changes(self):
        import time
        
        url = self.RUBYX_API+'/_changes?feed=continuous;include_docs=true;since='
        error = None
        seq = 'now'
        change_data = []
        last_update = 0
        
        while not error:
            with self.requests_session.get(url+seq, timeout=90, headers={"Authorization": self.auth}, stream=True) as resp:
                try:
                    for chunk in resp.iter_lines():
                        if chunk:  # filter out keep-alive new chunks
                            chunk = chunk.decode("utf-8")
                            changes = chunk.split('\n')  # make sure we handle individual changes
                            for change in changes:
                                print(change)
                                if(change.startswith('{')):
                                    data = json.loads(change)
                                    change_data.append(data)
                                    # handle changes in batches of 10 seconds
                                    # to avoid cluttering the HTTP pipeline
                                    if last_update < time.time() - 10:
                                        error, seq = self.handle_changes(change_data, seq)
                                        change_data = []
                                        last_update = time.time()
                except requests.exceptions.ChunkedEncodingError as e:
                    # this sometimes occurs with the new nginx proxy
                    # in front of the couchdb server when the response
                    # remains empty and the in-between proxy times out?
                    # luckily it seems we can just continue polling to keep
                    # this working as intended.
                    pass

    
    '''
    Handle a batch of changes by sending a notification if it
    concerns a new document.
    
    @todo trigger a list of agents when new inscope has been added
    @todo trigger actions for every new domain
    @todo trigger actions for every new ip
    @todo trigger actions for every new url
    '''
    def handle_changes(self, changes, seq):
        error = None
        source = ''
        message = ''

        # keep track of new documents to trigger hooks in bulk
        new = {'program': [], 'domain': [], 'ip': [], 'service': [], 'url': []}
        update = {'program': [], 'domain': [], 'ip': [], 'service': [], 'url': []}
        
        print('Looping changes')
        for change in changes:
            if 'error' in change:
                error = change['error']
            elif 'last_seq' in change:
                seq = change['last_seq']
            elif 'changes' in change:
                seq = change['seq']
                if change['changes'][-1]['rev'].startswith('1-'):  # check if it is the first revision of a document
                    
                    if 'type' in change['doc'] and change['doc']['type'] in new.keys():
                        new[change['doc']['type']].append(change['id'])
                    
                    # notify on new domain
                    if 'type' in change['doc'] and change['doc']['type'] == 'domain':
                        if 'source' in change['doc']:
                            message += '['+change['doc']['source']+'] '
                        message += change['id'] + ' was added\n'
                        
                    # also notify when a new notification was added
                    if 'type' in change['doc'] and change['doc']['type'] == 'alert':
                        message += '[ALERT] '
                        if 'source' in change['doc'] and change['doc']['source']:
                            message += '['+change['doc']['source']+'] '
                        message += change['doc']['message']+'\n'
                else:
                    # it's not new, hence it's an updated document:
                    if 'type' in change['doc'] and change['doc']['type'] in new.keys():
                        update[change['doc']['type']].append(change['id'])
        
        if message:
            if self.discord_webhook:
                # bypass Discord's 2000 character restriction
                # by breaking it up  and sending multiple messages
                # thanks to @0xJeti / closes #89
                for i in range(0,len(message),1000):
                    chunk = message[i:i+1000]
                    requests.post(self.discord_webhook, json.dumps({'content': chunk}), headers={'Content-Type': 'application/json'})

        for doctype in new.keys():
            if len(new[doctype]) > 0:
                self.execute_hooks('new', doctype, new[doctype])
        for doctype in update.keys():
            if len(update[doctype]) > 0:
                self.execute_hooks('update', doctype, update[doctype])
        
        return error, seq
    
    '''
    Execute hooks in ~/.rubyx/hooks/{domain,ip,url,service}/{new,update}/
    '''
    def execute_hooks(self, hooktype, doctype, identifiers):
        print('Executing '+hooktype+'_'+doctype)
        # first iteration without queueing: just
        # run the programs all at once in the background
        from subprocess import Popen
        from os import listdir
        from os.path import isfile, join, expanduser
        hookdir = expanduser('~/.rubyx/hooks/'+doctype+'/'+hooktype+'/')
        try:
            scripts = [join(hookdir, f) for f in listdir(hookdir) if isfile(join(hookdir, f)) and f.endswith('.sh')]
            for script in scripts:
                print('Running '+script+'...')
                args = [script]
                args.extend(identifiers)
                p = Popen(args)
        except FileNotFoundError:
            pass
        except Exception as e:
            print(e)
            # possibly the hooks directory doesn't exist
            pass
        #p = Popen([, 'ls'])
    
    '''
    Run an agent by triggering the Lambda HTTP endpoint for an agent
    
    @todo - should the invocation be run asynchronously?
    '''
    def run_agent(self, agent_name, program_name):
        r = self.requests_session.get(self.RUBYX_API+'/agents_api_gateway', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
        gateway = r.json()['url']
        
        return requests.get(gateway+agent_name+'?program='+requests.utils.quote(program_name, safe='')).text
    
    '''
    Configure the gateway URL
    '''
    def set_agent_gateway(self, url):
        # see if it exists
        gateway = self.get_document('agents_api_gateway')
        if gateway:
            self.update_document('config', 'agents_api_gateway', {'url': url})
        else:
            self.requests_session.put(self.RUBYX_API+'/agents_api_gateway', json.dumps({'doctype': 'config', 'url': url}), headers={"Authorization": self.auth})

    '''
    Register a new agent to the database
    '''
    def create_new_agent(self, agent_name):
        if not agent_name:
            return
        else:
            agent = {"type": "agent", "name": agent_name}
            r = self.requests_session.put(self.RUBYX_API+'/agent_'+agent_name, json.dumps(agent), headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
    
    '''
    Push an alert into the DB
    '''
    def create_alert(self, alert_message, program_name, source_name):
        if type(alert_message) is list:
            alert_message = '\n'.join(alert_message)
        alert = {
            "type": "alert",
            "program": program_name,
            "message": alert_message,
            "_id": alert_message,
            "source": source_name
        }
        r = self.requests_session.post(self.RUBYX_API, json.dumps(alert), headers={"Content-Type": "application/json", "Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
            
    '''
    List scopes across programs using _view/scope - does not support filtering by program.
    To get scope for one program, use `rubyx program scope`.
    '''
    def get_scopes(self, show_disabled=False):
        r = self.requests_session.get(
            self.RUBYX_API+'/_design/rubyx/_view/scope',
            headers={"Authorization": self.auth}
        )
        results = [ r for r in r.json()['rows'] if show_disabled or r['key'][0] ]
        return [r['key'][2] for r in results if r['key'][1].lower() == "in"], [r['key'][2] for r in results if r['key'][1].lower() == "out"]
        
    '''
    Get a list of documents based on a search term by tags, filtered by doctype
    '''
    def search_tags(self, key, value, doctype = None, program = None, show_disabled=False , show_empty_scope=False):
        if doctype and doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/search_tags?key=["'+requests.utils.quote(key, safe='')+'", "'+requests.utils.quote(value, safe='')+'"]', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
            
        results = r.json()['rows']
        
        # filter results based on the doctype and program name
        if doctype:
            results = [x for x in results if x['value'][0] == doctype]
        # restrict further if needs to match a specific program
        if not doctype == 'program' and program:
            results = [x for x in results if x['value'][2] == program]
        # remove disabled programs unles the --show-disabled flag is set
        if not show_disabled and doctype == 'program':
            active_programs = self.get_programs(show_disabled=show_disabled, show_empty_scope=show_empty_scope)
            results = [x for x in results if x['value'][2] in active_programs]
        
        return list(set([x['value'][1] for x in results if x['key'][0] == key]))
    

    def process_tags(self, tags, append_tags=False):
        '''
        Process a list of -t <name:value>... and return 
        as a dict of scalars or arrays.
        '''
        tagmap = {}
        for x in tags:
            k, v = x.split(':', 1)
            if k not in tagmap:
                tagmap[k] = v if not append_tags else [v]
            else:
                if type(tagmap[k]) is list:
                    tagmap[k].extend([v])
                else:
                    tagmap[k] = [v, tagmap[k]]
        return tagmap
    
    '''
    Get a list of documents based on a search term by tags, filtered by doctype
    '''
    def search_tags_between(self, key, value, before_after, doctype = None, program = None, show_disabled=False, show_empty_scope=False):
        if doctype and doctype not in self.doctypes:
            raise Exception('This doctype is not supported')
        if before_after == 'before':
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/search_tags?startkey=["'+requests.utils.quote(key, safe='')+'"]&endkey=["'+requests.utils.quote(key, safe='')+'","'+requests.utils.quote(value, safe='')+'"]', headers={"Authorization": self.auth})
        elif before_after == 'after':
            r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/search_tags?startkey=["'+requests.utils.quote(key, safe='')+'","'+requests.utils.quote(value, safe='')+'"]&endkey=["'+key+'ZZZZZ"]&', headers={"Authorization": self.auth})
        if 'error' in r.json():
            raise Exception('RUBYX server error: '+r.json()['error'])
            
        results = r.json()['rows']
        
        # filter results based on the doctype and program name
        if doctype:
            results = [x for x in results if x['value'][0] == doctype]
        if not doctype == 'program' and program:
            results = [x for x in results if x['value'][2] == program]
        if not show_disabled:
            active_programs = self.get_programs(show_disabled=show_disabled, show_empty_scope=show_empty_scope)
            results = [x for x in results if x['value'][2] in active_programs]
        
        return list(set([x['value'][1] for x in results if x['key'][0] == key]))
    
    def get_tags(self, tagname, program_name=None):
        filter_name = 'key="'+requests.utils.quote(tagname, safe='')+'"' if tagname else ''
        r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx/_view/tags?'+filter_name, headers={"Authorization": self.auth})
        results = r.json()['rows']
        
        if tagname:
            return [x['value'][0]+' '+x['value'][1] for x in results if not(not(program_name)) == (x['value'][2] == program_name)]
        else:
            return set([x['key'] for x in results if not(not(program_name)) == (x['value'][2] == program_name)])
    
    def server_upgrade(self, admin, password):
        '''
        Upgrade server to the latest views and ensure access rights are correct (see https://github.com/honoki/rubyx-server/pull/2)
        '''
        print('Downloading latest views from https://github.com/honoki/rubyx-server')
        r = self.requests_session.get('https://raw.githubusercontent.com/honoki/rubyx-server/main/couchdb/views.json')
        views = r.json()
        print('Comparing to current views...')
        admin_auth = 'Basic '+base64.b64encode((admin+':'+password).encode('utf-8')).decode('utf-8')
        r = self.requests_session.get(self.RUBYX_API+'/_design/rubyx', headers={"Authorization": admin_auth})
        if r.status_code == 401:
            print('[Error] Wrong administrator credentials provided.')
            return
        rev = r.json()['_rev']
        if not self.docs_are_equal(r.json(), views):
            print('Pushing views to server...')
            views['_rev'] = rev
            r = self.requests_session.put(self.RUBYX_API+'/_design/rubyx', json.dumps(views), headers={"Authorization": admin_auth, 'Content-Type': 'application/json'})
            if r.status_code == 201:
                print('Server upgrade complete, please allow a few minutes for all reindexing to complete. View the progress on '+self.RUBYX_API.replace('/rubyx','')+'/_utils/#/activetasks')
            else:
                print('Unexpected error: '+str(r.status_code))
                print(r.text)
        else:
            print('Server already up to date.')
        print('Validating user access rights...')
        r = self.requests_session.get(self.RUBYX_API+'/access-test') 
        if r.status_code == 401:
            print('No issues detected!')
        else:
            print('CRITICAL! Unauthorized access to the database is enabled. Fixing configuration...')
            r = self.requests_session.put(self.RUBYX_API+'/_security', '{"admins":{"names":[]},"members":{"names":["rubyx"]}}', headers={"Authorization": admin_auth, 'Content-Type': 'application/json'})
            if r.status_code == 200:
                print('Configuration corrected, all\'s good!')
            else:
                print('Something went wrong when trying to correct the misconfiguration. Please run the following to fix manually:')
                print('curl -ik -X PUT "'+self.RUBYX_API+'/_security" -u admin:password -d \'\'')
    
    def debug(self, msg):
        if self.do_debug:
            print('[DEBUG] '+msg)


import os, sys, json, pytest, io
#currentdir = os.path.dirname(os.path.realpath(__file__))
#parentdir = os.path.dirname(currentdir)
sys.path.append("..")
from src import RubyxClient

with open(os.path.expanduser('~/.rubyx/config-test.json')) as f:
    conf = json.load(f)

def rubyx(cmd):
    return RubyxClient(cmd, conf).run()

def list_equals(a, b):
    diff = set(a) ^ set(b)
    return not diff and len(a) == len(b)

'''
rubyx ( new | use | disable | enable ) <program> [ -t key:value ]...
rubyx programs [ --show-disabled --show-empty-scope ]
'''
def test_program():
    rubyx('new test')
    # program without scope is not going to show up
    assert 'test' not in rubyx('programs')
    assert 'test' in rubyx('programs --show-empty-scope')  
    
    rubyx('disable test')
    assert [] == rubyx('programs')
    assert 'test' not in rubyx('programs --show-empty-scope')
    assert 'test' in rubyx('programs --show-empty-scope --show-disabled')
    
    rubyx('enable test')
    assert 'test' not in rubyx('programs')
    assert 'test' in rubyx('programs --show-empty-scope')
    
    rubyx('new testtag -t test:tag -t test2:tag2')
    assert rubyx('program active') == 'testtag'
    assert json.loads(rubyx('show testtag'))['tags']['test'] == 'tag'
    assert json.loads(rubyx('show testtag'))['tags']['test2'] == 'tag2'

def test_program_special_chars(monkeypatch):
    rubyx('new test/weird&char?')
    # program without scope is not going to show up
    assert 'test/weird&char?' not in rubyx('programs')
    assert 'test/weird&char?' in rubyx('programs --show-empty-scope')
    rubyx('use test/weird&char?')
    rubyx('program update test/weird&char? -t test:tagupdated -t test2:tagupdated2')
    rubyx('inscope add *.weird.com')
    rubyx('domain add sub.weird.com')
    assert 'sub.weird.com' in rubyx('domains')
    # answer 'yes' to 'are you sure you want to remove'
    rubyx('domain remove sub.weird.com')
    rubyx('inscope remove *.weird.com')
    
    
def test_program_use():
    rubyx('use test')
    assert rubyx('program active') == 'test'
    rubyx('use testtag')
    assert rubyx('program active') == 'testtag'
    with pytest.raises(Exception):
        rubyx('use notexist')
    
'''
rubyx programs where <tag_name> is [ before | after ] <value>
'''
def test_programs_where():
    assert 'testtag' in rubyx('programs where test is tag --show-empty-scope')
    assert 'testtag' in rubyx('programs where test2 is tag2 --show-empty-scope')
    assert 'testtag' in rubyx('programs where test is before tagzzz --show-empty-scope')
    assert 'testtag' not in rubyx('programs where test is after tagZZZ --show-empty-scope')
    assert 'testtag' not in rubyx('programs where test is before aaatag --show-empty-scope')
    assert 'testtag' in rubyx('programs where test is after aaatag --show-empty-scope')
    
'''
rubyx program ( active | update ( <program>... | - ) -t key:value... [--append-tags])
'''
def test_program_update_tags():
    rubyx('program update testtag -t test:tagupdated -t test2:tagupdated2')
    assert json.loads(rubyx('show testtag'))['tags']['test'] == 'tagupdated'
    rubyx('program update testtag -t test:tagupdated -t test:testarray')
    assert list_equals(json.loads(rubyx('show testtag'))['tags']['test'], ['tagupdated','testarray'])
    rubyx('program update testtag -t test:tagone')
    rubyx('program update testtag -t test:tagtwo --append-tags')
    assert list_equals(json.loads(rubyx('show testtag'))['tags']['test'], ['tagone','tagtwo'])
    rubyx('program update testtag -t test:tagone -t test:tagtwo')
    rubyx('program update testtag -t test:tagthree --append-tags')
    assert list_equals(json.loads(rubyx('show testtag'))['tags']['test'], ['tagone','tagtwo','tagthree'])

'''
rubyx ( inscope | outscope ) ( add | remove ) ( - | <element>... ) [ -p <program> ]
rubyx scope filter ( in | out ) [ (--wildcard [--top] ) ] [ ( -p <program> ) | ( --all [--show-disabled] ) ]
'''
def test_scope():
    rubyx('use test')
    rubyx('inscope add *.example.com')  
    rubyx('inscope add *.sub.example.com *.dev.example.com')
    assert list_equals(rubyx('scope in'), ['*.example.com', '*.sub.example.com', '*.dev.example.com'])

    
    rubyx('inscope remove *.dev.example.com')
    assert list_equals(rubyx('scope in'), ['*.example.com', '*.sub.example.com'])
    assert list_equals(rubyx('scope in --wildcard'), ['example.com', 'sub.example.com'])
    assert list_equals(rubyx('scope in --wildcard --top'), ['example.com'])
    
    rubyx('inscope add *.example.co.uk -p testtag')  
    rubyx('inscope add *.sub.example.co.uk *.dev.example.co.uk -p testtag')
    assert list_equals(rubyx('scope in -p testtag'), ['*.example.co.uk', '*.sub.example.co.uk', '*.dev.example.co.uk'])
    rubyx('inscope remove *.dev.example.co.uk -p testtag')
    assert list_equals(rubyx('scope in -p testtag'), ['*.example.co.uk', '*.sub.example.co.uk'])
    
    rubyx('outscope add *.dev.example.com')
    assert rubyx('scope out') == ['*.dev.example.com']
    assert rubyx('scope out --wildcard') == ['dev.example.com']
    assert rubyx('scope out --wildcard --top') == ['dev.example.com']
    rubyx('outscope add sub.dev.example.com')
    assert rubyx('scope out --wildcard --top') == ['dev.example.com']
    
    assert list_equals(rubyx('scope in --all'), [
        '*.example.com',
        '*.sub.example.com',
        '*.example.co.uk',
        '*.sub.example.co.uk',
    ])
    
    rubyx('disable testtag')
    assert list_equals(rubyx('scope in --all'), [
        '*.example.com',
        '*.sub.example.com',
    ])
    # TODO: this is inconsistent behaviour, and returns ONLY
    # disabled programs. See https://github.com/honoki/rubyx-client/issues/47
    assert list_equals(rubyx('scope in --all --show-disabled'), [
        '*.example.com',
        '*.sub.example.com',
        '*.example.co.uk',
        '*.sub.example.co.uk',
    ])
    rubyx('enable testtag')

    # test URL scopes
    rubyx('inscope add http://url.example.com/ https://url2.example.com')
    assert 'url.example.com' in rubyx('scope in')
    assert 'url2.example.com' in rubyx('scope in')
    rubyx('outscope add http://URL3.EXaMPLe.cOM/')
    assert 'url3.example.com' in rubyx('scope out')
    
def test_scope_filter(monkeypatch):
    rubyx('use test')
    monkeypatch.setattr('sys.stdin', io.StringIO('one.example.com\ntwo.example.com\nsub.dev.example.com'))
    assert list_equals(rubyx('scope filter in'), ['one.example.com', 'two.example.com'])
    monkeypatch.setattr('sys.stdin', io.StringIO('one.example.com\ntwo.example.com\nsub.dev.example.com'))
    assert list_equals(rubyx('scope filter out'), ['sub.dev.example.com'])
    
    monkeypatch.setattr('sys.stdin', io.StringIO('''
one.example.com
two.example.com
sub.dev.example.com
one.example.co.uk
two.example.co.uk
sub.dev.example.co.uk
'''))
    assert list_equals(rubyx('scope filter in --all'), ['one.example.com', 'two.example.com','one.example.co.uk', 'two.example.co.uk', 'sub.dev.example.co.uk'])
    monkeypatch.setattr('sys.stdin', io.StringIO('''
one.example.com
two.example.com
sub.dev.example.com
one.example.co.uk
two.example.co.uk
sub.dev.example.co.uk
'''))
    assert list_equals(rubyx('scope filter out --all'), ['sub.dev.example.com'])
    
'''
rubyx domain ( add | remove | update ) ( - | <domain>... ) [ -p <program> -s <source> --show-new ( -t key:value... [--append-tags] ) ]
rubyx domains [ --view <view> ( -p <program> | ( --all [--show-disabled] ) ) ]
'''
def test_domains(monkeypatch):
    assert rubyx('domains') == []
    
    # test adding and tagging
    rubyx('domain add one.example.com two.example.com three.example.com')
    assert list_equals(rubyx('domains'), ['one.example.com','two.example.com','three.example.com'])
    rubyx('domain add four.example.com:4.4.4.4')
    assert list_equals(rubyx('domains'), ['one.example.com','two.example.com','three.example.com', 'four.example.com'])
    rubyx('domain update one.example.com -t tagging:test_domains')
    assert json.loads(rubyx('show one.example.com'))['tags']['tagging'] == 'test_domains'
    rubyx('domain update one.example.com -t tagging:test_domains -t tagging:array')
    assert list_equals(json.loads(rubyx('show one.example.com'))['tags']['tagging'], ['test_domains', 'array'])
    rubyx('domain update one.example.com -t tagging:append --append-tags')
    assert list_equals(json.loads(rubyx('show one.example.com'))['tags']['tagging'], ['test_domains', 'array', 'append'])
    rubyx('domain update one.example.com -t tagging:overwrite')
    assert json.loads(rubyx('show one.example.com'))['tags']['tagging'] == 'overwrite'
    
    # test domains through input pipe
    monkeypatch.setattr('sys.stdin', io.StringIO('''
pipe1.example.com
pipe2.example.com
pipe3.dev.example.com
'''))
    rubyx('domain add -')
    assert list_equals(rubyx('domains'), ['one.example.com','two.example.com','three.example.com', 'four.example.com', 'pipe1.example.com', 'pipe2.example.com'])
    monkeypatch.setattr('sys.stdin', io.StringIO('''
pipe1.example.com
pipe2.example.com
pipe3.dev.example.com
'''))
    rubyx('domain remove -')
    assert list_equals(rubyx('domains'), ['one.example.com','two.example.com','three.example.com', 'four.example.com'])

    # test adding out-of-scope domains
    rubyx('domain add not-in-scope.example.be')
    assert 'not-in-scope.example.be' not in rubyx('domains')
    rubyx('domain add not-in-scope.example.be --ignore-scope')
    assert 'not-in-scope.example.be' in rubyx('domains')
    rubyx('domain remove not-in-scope.example.be')
    
    # test ips
    assert list_equals(json.loads(rubyx('show four.example.com'))['ips'], ['4.4.4.4'])
    rubyx('domain update four.example.com:4.4.4.4,4.4.4.5,4.4.4.6')
    assert list_equals(json.loads(rubyx('show four.example.com'))['ips'], ['4.4.4.4', '4.4.4.5','4.4.4.6'])
    rubyx('domain update four.example.com:4.4.4.4,4.4.4.5,4.4.4.6 four.example.com:4.4.4.4,4.4.4.7 -s pytest')
    assert list_equals(json.loads(rubyx('show four.example.com'))['ips'], ['4.4.4.4', '4.4.4.5','4.4.4.6','4.4.4.7'])
    
    # test source
    assert json.loads(rubyx('show four.example.com'))['source'] == 'pytest'
    
    # test --show-new for new and updated domains
    assert rubyx('domain add five.example.com --show-new -s pytest') == ['[NEW] five.example.com']
    assert json.loads(rubyx('show five.example.com'))['source'] == 'pytest'
    assert rubyx('domain update five.example.com -n') == None
    assert rubyx('domain update five.example.com:5.5.5.5 -n') == ['[UPDATED] five.example.com']
    assert rubyx('domain update five.example.com:5.5.5.5 -n') == None
    assert rubyx('domain update five.example.com:5.5.5.5 -n -s updated') == ['[UPDATED] five.example.com']
    
    # test remove
    rubyx('domain remove four.example.com five.example.com')
    assert list_equals(rubyx('domains'), ['one.example.com','two.example.com','three.example.com'])
    
    

'''
rubyx domains where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
'''
def test_domains_where():
    assert list_equals(rubyx('domains'), ['one.example.com','two.example.com','three.example.com'])
    rubyx('domain add four.example.com:4.4.4.4')
    
    assert list_equals(rubyx('domains where ip is 4.4.4.4'), ['four.example.com'])
    assert 'four.example.com' in rubyx('domains where ip is before 4.4.4.4ZZZ')
    assert 'four.example.com' not in rubyx('domains where ip is before 1.1.1.1')
    assert 'four.example.com' in rubyx('domains where ip is after 4.4.4.0')
    assert 'four.example.com' not in rubyx('domains where ip is after 4.4.4.9')
    
    
    assert list_equals(rubyx('domains where tagging is overwrite'), ['one.example.com'])
    assert 'one.example.com' in rubyx('domains where tagging is before overwriteZZZ')
    assert 'one.example.com' not in rubyx('domains where tagging is before overwritA')
    assert 'one.example.com' in rubyx('domains where tagging is after overwrita')
    assert 'one.example.com' not in rubyx('domains where tagging is after overwritZ')

def test_domains_underscore():
    rubyx('domain add _one.example.com _two.example.com')
    assert '_one.example.com' in rubyx('domains')
    assert '_two.example.com' in rubyx('domains')
    assert rubyx('domain add _three.example.com:3.0.3.0 -n') == ['[NEW] _three.example.com']
    assert rubyx('domains where ip is 3.0.3.0') == ['_three.example.com']
    assert rubyx('domain update _three.example.com:4.0.4.0 -n') == ['[UPDATED] _three.example.com']
    assert rubyx('domain remove _three.example.com -n') == ['[DELETED] _three.example.com']
    rubyx('domain remove _one.example.com _two.example.com')
    assert '_one.example.com' not in rubyx('domains')
    assert '_two.example.com' not in rubyx('domains')

def test_domains_resolved():
    rubyx('domain add one.example.com:1.1.1.1 two.example.com:2.2.2.2 three.example.com:127.0.0.1 four.example.com:10.0.0.1 five.example.com:192.168.0.1 six.example.com:172.16.0.0 seven.example.com:172.10.1.1')
    rubyx('domain update one.example.com:1.1.1.1 four.example.com:10.0.0.1')
    assert 'one.example.com' in rubyx('domains --resolved')
    assert 'one.example.com' in rubyx('domains --resolved --no-private')
    assert 'three.example.com' not in rubyx('domains --resolved --no-private')
    assert 'four.example.com' not in rubyx('domains --resolved --no-private')
    assert 'five.example.com' not in rubyx('domains --resolved --no-private')
    assert 'six.example.com' not in rubyx('domains --resolved --no-private')
    assert 'seven.example.com' in rubyx('domains --resolved --no-private')
    rubyx('domain add eight.example.com')
    assert 'eight.example.com' in rubyx('domains --unresolved')
    rubyx('domain remove two.example.com three.example.com four.example.com five.example.com six.example.com seven.example.com eight.example.com')

'''
rubyx ips [ --filter-cdns ( -p <program> | ( --all [--show-disabled] ) ) ]
'''
def test_ips(monkeypatch):
    assert rubyx('ips') == []
    # test adding and tagging
    rubyx('ip add 1.1.1.1 2.2.2.2 3.3.3.3')
    assert list_equals(rubyx('ips'), ['1.1.1.1','2.2.2.2','3.3.3.3'])
    rubyx('ip add 4.4.4.4:four.example.com')
    assert list_equals(rubyx('ips'), ['1.1.1.1','2.2.2.2','3.3.3.3', '4.4.4.4'])
    rubyx('ip update 1.1.1.1 -t tagging:test_domains')
    assert json.loads(rubyx('show 1.1.1.1'))['tags']['tagging'] == 'test_domains'
    rubyx('ip update 1.1.1.1 -t tagging:test_domains -t tagging:array')
    assert list_equals(json.loads(rubyx('show 1.1.1.1'))['tags']['tagging'], ['test_domains', 'array'])
    rubyx('ip update 1.1.1.1 -t tagging:append --append-tags')
    assert list_equals(json.loads(rubyx('show 1.1.1.1'))['tags']['tagging'], ['test_domains', 'array', 'append'])
    rubyx('ip update 1.1.1.1 -t tagging:overwrite')
    assert json.loads(rubyx('show 1.1.1.1'))['tags']['tagging'] == 'overwrite'
    
    # test ips through input pipe
    monkeypatch.setattr('sys.stdin', io.StringIO('''
11.11.11.11
22.22.22.22
33.33.33.33
'''))
    rubyx('ip add -')
    assert list_equals(rubyx('ips'), ['1.1.1.1','2.2.2.2','3.3.3.3', '4.4.4.4', '11.11.11.11', '22.22.22.22', '33.33.33.33'])
    monkeypatch.setattr('sys.stdin', io.StringIO('''
11.11.11.11
22.22.22.22
33.33.33.33
'''))
    rubyx('ip remove -')
    assert list_equals(rubyx('ips'), ['1.1.1.1','2.2.2.2','3.3.3.3','4.4.4.4'])
    
    # test domains
    assert list_equals(json.loads(rubyx('show 4.4.4.4'))['domains'], ['four.example.com'])
    rubyx('ip update 4.4.4.4:four.example.com,four2.example.com,four3.example.com')
    assert list_equals(json.loads(rubyx('show 4.4.4.4'))['domains'], ['four.example.com', 'four2.example.com','four3.example.com'])
    rubyx('ip update 4.4.4.4:four4.example.com,four5.example.com 4.4.4.4:four6.example.com,four7.example.com -s pytest')
    assert list_equals(json.loads(rubyx('show 4.4.4.4'))['domains'], ['four.example.com', 'four2.example.com','four3.example.com','four4.example.com', 'four5.example.com','four6.example.com','four7.example.com'])
    
    # test source
    assert json.loads(rubyx('show 4.4.4.4'))['source'] == 'pytest'
    
    # test --show-new for new and updated domains
    assert rubyx('ip add 5.5.5.5 --show-new -s pytest') == ['[NEW] 5.5.5.5']
    assert json.loads(rubyx('show 5.5.5.5'))['source'] == 'pytest'
    assert rubyx('ip update 5.5.5.5 -n') == None
    assert rubyx('ip update 5.5.5.5:five.example.com -n') == ['[UPDATED] 5.5.5.5']
    assert rubyx('ip update 5.5.5.5:five.example.com -n') == None
    assert rubyx('ip update 5.5.5.5:five.example.com -n -s updated') == ['[UPDATED] 5.5.5.5']
    
    # test remove
    rubyx('ip remove 4.4.4.4 5.5.5.5')
    assert list_equals(rubyx('ips'), ['1.1.1.1','2.2.2.2','3.3.3.3'])

def test_cidr_scope(monkeypatch):
    rubyx('inscope add 3.2.1.0/23')
    assert '3.2.1.0/23' in rubyx('scope in')
    rubyx('outscope add 3.2.1.254')
    assert '3.2.1.254' in rubyx('scope out')
    # at the moment this only impacts URLs
    rubyx('url add http://1.2.3.4:80 http://3.2.1.1:80')
    assert 'http://3.2.1.1:80' in rubyx('urls')
    assert 'http://1.2.3.4:80' not in rubyx('urls')
    rubyx('url remove http://3.2.1.1:80')
    # ensure the URL is added if --ignore-scope is used
    rubyx('url add http://1.2.3.4:80 http://3.2.1.1:80 --ignore-scope')
    assert 'http://3.2.1.1:80' in rubyx('urls')
    assert 'http://1.2.3.4:80' in rubyx('urls')
    rubyx('url remove http://3.2.1.1:80 http://1.2.3.4:80')
    

'''
rubyx ips where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
'''
def test_ips_where():
    assert list_equals(rubyx('ips'), ['1.1.1.1','2.2.2.2','3.3.3.3'])
    rubyx('ip add 4.4.4.4:four.example.com')
    
    assert list_equals(rubyx('ips where domain is four.example.com'), ['4.4.4.4'])
    assert '4.4.4.4' in rubyx('ips where domain is before four.example.coZ')
    assert '4.4.4.4' not in rubyx('ips where domain is before AAAA.example.com')
    assert '4.4.4.4' in rubyx('ips where domain is after four.example.aaa')
    assert '4.4.4.4' not in rubyx('ips where domain is after four.example.ZZZ')
    
    
    assert list_equals(rubyx('ips where tagging is overwrite'), ['1.1.1.1'])
    assert '1.1.1.1' in rubyx('ips where tagging is before overwriteZZZ')
    assert '1.1.1.1' not in rubyx('ips where tagging is before overwritA')
    assert '1.1.1.1' in rubyx('ips where tagging is after overwrita')
    assert '1.1.1.1' not in rubyx('ips where tagging is after overwritZ')

'''
rubyx urls [ -d <hostname> | ( -p <program> | ( --all [--show-disabled] ) ) ] [--with-query]
'''
def test_urls():
    assert list_equals(rubyx('urls'), [])
    
    # plain http
    rubyx('url add http://one.example.com/one')
    rubyx('url add http://two.example.com/two http://three.example.com/three')
    assert list_equals(rubyx('urls'), [
        'http://one.example.com/one',
        'http://two.example.com/two',
        'http://three.example.com/three'
    ])
    
    # https
    rubyx('url add http://three.example.com/1 http://three.example.com/2 http://three.example.com/3 -t protocol:http')
    rubyx('url add https://three.example.com/1 https://three.example.com/2 https://three.example.com/3 -t protocol:https')
    # with query strings
    rubyx('url add http://three.example.com/query?one=two&three=four')
    rubyx('url add http://three.example.com/query?five=six')
    rubyx('url add https://three.example.com:8080/url?some=what')
    # relative path with specified domain
    rubyx('url add /relative -d three.example.com')
    
    # when port :80 or :443 are explicitly set in the URL, they need to be removed
    # so we don't duplicate value where port is not set!
    rubyx('url add http://three.example.com:80/1 https://port.example.com:443/port https://port.example.com/port')
    
    assert list_equals(rubyx('urls'), [
        'http://one.example.com/one',
        'http://two.example.com/two',
        'http://three.example.com/three',
        'http://three.example.com/1',
        'http://three.example.com/2',
        'http://three.example.com/3',
        #https
        'https://three.example.com/1',
        'https://three.example.com/2',
        'https://three.example.com/3',
        #querystrings
        'http://three.example.com/query',
        'https://three.example.com:8080/url',
        #relative
        'http://three.example.com/relative',
        
        # explicit ports:
        # https://port.example.com:443/port should not be listed
        # http://three.example.com:80/1 should not be listed
        'https://port.example.com/port'
    ])
    
    # including query strings
    assert list_equals(rubyx('urls -q'), [
        'http://one.example.com/one',
        'http://two.example.com/two',
        'http://three.example.com/three',
        'http://three.example.com/1',
        'http://three.example.com/2',
        'http://three.example.com/3',
        #https
        'https://three.example.com/1',
        'https://three.example.com/2',
        'https://three.example.com/3',
        #querystrings
        'http://three.example.com/query?one=two&three=four',
        'http://three.example.com/query?five=six',
        'https://three.example.com:8080/url?some=what',
        #relative
        'http://three.example.com/relative',
        # explicit ports:
        'https://port.example.com/port'
    ])
    
    # including status codes and content lengths
    rubyx('url add http://three.example.com/c 200 1234')
    assert 'http://three.example.com/c' in rubyx('urls')
    
'''
rubyx urls where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
'''
def test_urls_where():
    assert 'http://one.example.com/one' in rubyx('urls where hostname is one.example.com')
    assert 'https://three.example.com:8080/url' in rubyx('urls where port is 8080')
    
    assert 'http://three.example.com/1' in rubyx('urls where protocol is http')
    assert 'http://three.example.com/1' in rubyx('urls where protocol is after httA')
    assert 'http://three.example.com/1' not in rubyx('urls where protocol is after httpZZZ')
    assert 'http://three.example.com/1' in rubyx('urls where protocol is before httZ')
    assert 'http://three.example.com/1' not in rubyx('urls where protocol is before httA')

'''
rubyx url remove ( - | <url>... )
'''
def test_urls_remove(monkeypatch):
    rubyx('url remove https://three.example.com:8080/url')
    assert 'https://three.example.com:8080/url' not in rubyx('urls')
    monkeypatch.setattr('sys.stdin', io.StringIO('''
http://one.example.com/one
http://two.example.com/two
http://three.example.com/three
http://three.example.com/1
http://three.example.com/2
http://three.example.com/3
http://three.example.com/c
'''))
    rubyx('url remove -')
    assert list_equals(rubyx('urls'), [
        #https
        'https://three.example.com/1',
        'https://three.example.com/2',
        'https://three.example.com/3',
        #querystrings
        'http://three.example.com/query',
        #relative
        'http://three.example.com/relative',
        # explicit ports:
        'https://port.example.com/port'
    ])

'''
rubyx services [ -p <program> | ( --all [--show-disabled] ) ]
rubyx service add ( - | <service>... ) [ -s <source> -p <program> --show-new ( -t key:value... [ --append-tags ] ) ]
'''
def test_services():
    assert rubyx('services') == []
    rubyx('service add 1.1.1.1:11 2.2.2.2:22:ssh 3.3.3.3:33')
    assert rubyx('service add 3.3.3.3:33 -n') == None
    assert rubyx('service add 3.3.3.3:33:some -n') == ['[UPDATED] 3.3.3.3:33']
    assert rubyx('service add 3.3.3.3:33:some -n') == None
    assert rubyx('service add 3.3.3.3:33:some -n -s updated') == ['[UPDATED] 3.3.3.3:33']
    
    assert list_equals(rubyx('services'), [
        '1.1.1.1:11',
        '2.2.2.2:22',
        '3.3.3.3:33'
    ])
    
    rubyx('service add 1.1.1.1:11 -t tagging:test_services')
    assert json.loads(rubyx('show 1.1.1.1:11'))['tags']['tagging'] == 'test_services'
    rubyx('service add 1.1.1.1:11 -t tagging:test_services -t tagging:array')
    assert list_equals(json.loads(rubyx('show 1.1.1.1:11'))['tags']['tagging'], ['test_services', 'array'])
    rubyx('service add 1.1.1.1:11 -t tagging:append --append-tags')
    assert list_equals(json.loads(rubyx('show 1.1.1.1:11'))['tags']['tagging'], ['test_services', 'array', 'append'])
    rubyx('service add 1.1.1.1:11 -t tagging:overwrite')
    assert json.loads(rubyx('show 1.1.1.1:11'))['tags']['tagging'] == 'overwrite'

'''
rubyx services where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
'''
def test_services_where():
    assert '1.1.1.1:11' in rubyx('services where ip is 1.1.1.1')
    assert '1.1.1.1:11' in rubyx('services where port is 11')
    assert '2.2.2.2:22' in rubyx('services where service is ssh')
    assert '3.3.3.3:33' in rubyx('services where service is some')
    
    assert '1.1.1.1:11' in rubyx('services where tagging is overwrite')
    assert '1.1.1.1:11' in rubyx('services where tagging is after overwritA')
    assert '1.1.1.1:11' not in rubyx('services where tagging is after overwriteZZZ')
    assert '1.1.1.1:11' in rubyx('services where tagging is before overwritZ')
    assert '1.1.1.1:11' not in rubyx('services where tagging is before overwritA')

'''
rubyx service remove ( - | <service>... )
'''
def test_service_remove(monkeypatch):
    rubyx('service remove 1.1.1.1:11')
    assert '1.1.1.1:11' not in rubyx('services')
    monkeypatch.setattr('sys.stdin', io.StringIO('''
1.1.1.1:11
1.1.1.1:11
2.2.2.2:22
3.3.3.3:33
'''))
    rubyx('service remove -')
    assert list_equals(rubyx('services'), [])

'''
rubyx blacklist ( add | remove ) ( - | <element>... ) [ -p <program> ]
'''
def test_blacklist():
    rubyx('use test')
    rubyx('blacklist add blacklist.example.com')
    rubyx('domain add blacklist.example.com')
    assert 'blacklist.example.com' not in rubyx('domains')
    
    rubyx('blacklist remove blacklist.example.com')
    rubyx('domain add blacklist.example.com')
    assert 'blacklist.example.com' in rubyx('domains')
    
    rubyx('blacklist add 9.9.9.9')
    rubyx('domain add blacklistip.example.com:9.9.9.9')
    assert 'blacklistip.example.com' not in rubyx('domains')

'''
rubyx agent ( list | ( register | remove ) <agent>... | gateway [ <url> ] )
'''
def test_agent():
    assert rubyx('agent list') == []
    rubyx('agent register testagent')
    assert rubyx('agent list') == ['testagent']
    rubyx('agent register testagent2')
    rubyx('agent remove testagent')
    assert rubyx('agent list') == ['testagent2']
    rubyx('agent gateway http://localhost/gateway-test')
    assert rubyx('agent gateway') == 'http://localhost/gateway-test'
    
'''
rubyx agents
'''
def test_agents():
    assert rubyx('agents') == ['testagent2']

'''
rubyx run <agent> [ -p <program> ]
'''
pass

'''
rubyx listen
'''
pass

'''
rubyx alert ( - | <message> ) [ -s <source> ]
'''
pass

'''
rubyx tags [<name>] [ -p <program> | --all ]
'''
def test_tags():
    assert list_equals(rubyx('tags'), [
        'tagging',
        'protocol'
    ])
    
    assert list_equals(rubyx('tags --all'), [
        'test', # tag on program test
        'test2', # tag on proram test
        'tagging',
        'protocol'
    ])
    
    assert list_equals(rubyx('tags protocol'), [
        'https://three.example.com/1 https',
        'https://three.example.com/2 https',
        'https://three.example.com/3 https',
    ])
    assert list_equals(rubyx('tags tagging'), [
        '1.1.1.1 overwrite',
        'one.example.com overwrite',
    ])

'''
rubyx server upgrade
'''
def test_upgrade():
    # not yet sure how to test this here
    pass


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rubyx",
    version="1.0.0",
    author="@aituglo",
    author_email="contact@aituglo.com",
    description="Rubyx",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aituglo/rubyx",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['docopt', 'requests'],
    python_requires='>=3.5',
    entry_points = {
        'console_scripts': ['rubyx=src.rubyx:main'],
    }
)



#!/usr/bin/python3

"""Rubyx Client

Usage:
  rubyx ( new | use | disable | enable | rm ) <program> [ -t key:value ]...
  rubyx programs [ --show-disabled --show-empty-scope ]
  rubyx programs where <tag_name> is [ before | after ] <value> [ --show-disabled --show-empty-scope ]
  rubyx program ( active | update ( <program>... | - ) -t key:value... [--append-tags])
  rubyx domains [ --resolved [ --no-private ] | --unresolved | --view <view> ] [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx domains where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx domains where <tag_name> is [ before | after ] <value> ( and <tag_name> is [ before | after ] <value> )... [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx domain ( add | remove | update ) ( - | <domain>... ) [ -p <program> -s <source> --show-new ( -t key:value... [--append-tags] ) --ignore-scope ]
  rubyx ips [ --filter-cdns ( -p <program> | ( --all [--show-disabled] ) ) ]
  rubyx ips where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx ips where <tag_name> is [ before | after ] <value> ( and <tag_name> is [ before | after ] <value> )... [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx ip ( add | remove | update ) ( - | <ip>... ) [ -p <program> -s <source> --show-new ( -t key:value... [--append-tags] ) --ignore-scope ]
  rubyx scope ( in | out ) [ (--wildcard [--top] ) ] [ ( -p <program> ) | ( --all [--show-disabled] ) ]
  rubyx scope filter ( in | out ) [ (--wildcard [--top] ) ] [ ( -p <program> ) | ( --all [--show-disabled] ) ]
  rubyx ( inscope | outscope ) ( add | remove ) ( - | <element>... ) [ -p <program> ]
  rubyx urls [ -d <hostname> | ( -p <program> | ( --all [--show-disabled] ) ) ] [ --with-query | --root ]
  rubyx urls where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx urls where <tag_name> is [ before | after ] <value> ( and <tag_name> is [ before | after ] <value> )... [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx url add ( - | <url>... ) [ -d <hostname> -s <source> -p <program> --show-new ( -t key:value... [--append-tags] ) --ignore-scope ]
  rubyx url remove ( - | <url>... )
  rubyx services [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx services where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx services where <tag_name> is [ before | after ] <value> ( and <tag_name> is [ before | after ] <value> )... [ -p <program> | ( --all [--show-disabled] ) ]
  rubyx service add ( - | <service>... ) [ -s <source> -p <program> --show-new ( -t key:value... [ --append-tags ] ) --ignore-scope ]
  rubyx service remove ( - | <service>... )
  rubyx blacklist ( add | remove ) ( - | <element>... ) [ -p <program> ]
  rubyx agents
  rubyx agent ( list | ( register | remove ) <agent>... | gateway [ <url> ] )
  rubyx run <agent> [ -p <program> ]
  rubyx show ( - | <document>... )
  rubyx remove ( - | <document>... ) [ --yes ]
  rubyx listen
  rubyx alert ( - | <message>... ) [ -s <source> --show-new -t key:value... ] 
  rubyx tags [<name>] [ -p <program> | --all ] 
  rubyx server upgrade [--yes]

Options:
  -h --help            Show this screen.
  -p <program>         Select a program to limit the command to. Not required when the command "use" has been run before.
  -t key:value         Specify one or more custom tags to add to your document. Format as key:value
  -s <source>          Provide an optional source string to store information about the source of the modified data.
  -v --version         Show the program version
  -d <hostname>        Explicitly specify the hostname of a URL in case of relative paths
  -n, --show-new       Print new unique values that were added to the database, and didn't already exist
  -A, --all            Specify to get information across all programs. Incompatible with the -p flag
  -a, --append-tags    If a tag with the same name already exists on the document, make an array and append the new value(s)
  -q, --with-query     When listing URLs, show all URLs including queries
  -r, --root           When listing URLs, only list the roots of the URLs
  -w, --show-disabled  Combine with the flag --all/-A to include documents of disabled programs too
  -R, --resolved       When listing domains, only show resolved domain 
  -u, --unresolved     When listing domains, only show unresolved domains
  -x, --no-private     Combine with --resolved/-R, only show domains that don't resolve to a private IP address
  -y, --yes            Don't prompt for confirmation when deleting document or upgrading server
  -f, --ignore-scope   Ignore the scope (i.e. force) when adding a domain, url, ip or service
"""

import os
import sys
import json
import re
from . import rubyx_api
from urllib.parse import urlparse
from docopt import docopt

CONFIG_FILE = '~/.rubyx/config.json'
# Thanks https://regexr.com/3au3g
REGEX_DOMAIN = re.compile('^(?:[a-z0-9_](?:[a-z0-9-_]{0,61}[a-z0-9])?\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$')
# regex to match IP addresses and CIDR ranges - thanks https://www.regextester.com/93987
REGEX_IP = re.compile('^([0-9]{1,3}\\.){3}[0-9]{1,3}(/([0-9]|[1-2][0-9]|3[0-2]))?$')
VERSION = '1.3.2'

class RubyxClient:
    config = {}
    file_config = None
    arguments = None
    api = None
    
    def __init__(self, arguments, config=None):
        
        if not str(type(arguments)) == "<class 'docopt.Dict'>":
            self.arguments = docopt(__doc__, argv=arguments, version=VERSION)
        else:
            self.arguments = arguments
            
        if not config:
            self.load_config()
            self.file_config = True
        else:
            self.config = config
            self.file_config = False
        
        if 'username' not in self.config:
            exit('[ERROR] Required configuration was not found: username')
        elif 'password' not in self.config:
            exit('[ERROR] Required configuration was not found: password')
        elif 'couchdb' not in self.config:
            exit('[ERROR] Required configuration was not found: couchdb')
        
        else:
            self.api = rubyx_api.RUBYXApi(
                self.config['couchdb'],
                self.config['username'],
                self.config['password'],
                discord_webhook = self.config['discord_webhook'] if 'discord_webhook' in self.config else None,
                ignore_ssl_errors = self.config['ignore_ssl_errors'] if 'ignore_ssl_errors' in self.config else None,
                debug = self.config['debug'] if 'debug' in self.config else False
            )

    def new_program(self):
        # First set the program as the active one
        self.use_program(False)
         # If the program name seems to be a domain name, issue a warning
        # that this is not the ideal formatting
        if REGEX_DOMAIN.match(self.get_program()):
            print('[WARNING] You are adding a program name that looks like a domain name, which is discouraged. You will now not be able to add the domain '+self.get_program()+' to your database')
        # and add it to the db
        self.api.create_new_program(self.get_program(), tags=self.arguments['-t'])

    def list_programs(self, show_disabled, show_empty_scope):
        return self.api.get_programs(show_disabled=show_disabled, show_empty_scope=show_empty_scope)
    
    # updating programs this way only supports tags for now
    def update_programs(self, programs):
        update_programs = {}
            
        for program in programs:
            
            if(self.arguments['-t'] and len(self.arguments['-t']) > 0):
                update_programs[program] = {}
                update_programs[program]['tags'] = self.api.process_tags(self.arguments['-t'])
            
        updated = self.api.update_documents('program', {program: update_programs[program] for program in update_programs.keys()}, append_tags=self.arguments['--append-tags'])
    
    def list_agents(self):
        return [r['key'] for r in self.api.get_agents()]
    
    def register_agents(self, agent_names):
        for agent in agent_names:
            try:
                self.api.create_new_agent(agent)
            except:
                # this agent probably exists already
                print('Error adding new agent ' + agent + '...')
    
    def remove_agents(self, agent_names):
        for agent in agent_names:
            try:
                self.api.remove_document('agent', {'key': agent})
            except:
                continue
    
    '''
    Specify a program to be used, and avoid having to type -p <name>
    for every command. The config is stored in CONFIG_FILE.
    '''
    def use_program(self, check_exists=True):
        pro = self.arguments['<program>']
        if type(pro) is list:
            pro = pro[0]
        if check_exists and pro not in self.list_programs(True, True):
            raise Exception('This program does not exist.')
        self.config['program'] = pro
    
    '''
    Check whether an IP belongs to a CIDR range
    '''
    @staticmethod
    def ip_in_cidr(ip, cidr):
        from ipaddress import ip_network, ip_address
        return ip_address(ip) in ip_network(cidr, False)

    '''
    Make abstraction of where the program comes from. It should be either
    specified via the "use" command, or otherwise provided as the -p argument.
    '''
    def get_program(self):
        if self.arguments['-p']:
            return self.arguments['-p']
        elif 'program' in self.config:
            return self.config['program']
        else:
            raise Exception('You need to select a program to execute this action.')
 
    def get_scope(self):   
        if not self.arguments['--all']:
            if self.arguments['in']:
                (_, scope, _) = self.api.get_program_scope(self.get_program())
            elif self.arguments['out']:
                (_, _, scope) = self.api.get_program_scope(self.get_program())
        else: # get scope across all programs, making use of _view/scope
            inscope, outscope = self.api.get_scopes(show_disabled = self.arguments['--show-disabled'])
            if self.arguments['in']:
                scope = inscope
            elif self.arguments['out']:
                scope = outscope

        # filter the scope if --wildcard and/or --top flags are set
        if(self.arguments['--wildcard']):
            r_scope = []
            for s in scope:
                if s.startswith('*.'):
                    r_scope.append(s[2:])
            # only return top-level scope,
            # i.e. 
            if(self.arguments['--top']):
                all_scope = r_scope[:]  # make a copy of the scope before editing it
                for s in all_scope:  # run through each of the wildcard domains
                    for t in r_scope: # and compare to our limited scope copy
                        if s != t and t.endswith(s):
                            r_scope.remove(t)
                        elif s != t and s.endswith(t):
                            if s in r_scope:
                                r_scope.remove(s)

            return r_scope
        else:
            return scope

    def filter_scope(self):
        '''
        Compare stdin to the defined scope of the program and return the values that are considered in or out of scope
        '''
        if not self.arguments['--all']:
            (_, inscope, outscope) = self.api.get_program_scope(self.get_program())
        else:
            inscope, outscope = self.api.get_scopes()
        
        if self.arguments['in']:
            # if filtering against the inscope, we also need to ensure it's NOT in the outscope to avoid confusion
            match = ( line for line in process_stdin() if self.matches_scope(line, inscope) )
            return [ line for line in match if not self.matches_scope(line, outscope) ]
        else:
            # otherwise, print everything that matches the outscope + everything that does not match the inscope
            return [ line for line in process_stdin() if self.matches_scope(line, outscope) or not self.matches_scope(line, inscope) ]

    '''
    The RUBYX client is responsible for ensuring the added domain is not explicitly outscoped,
    and conforms to the expected format of a domain.
    
    Expected format:
        domains = [
         "a.example.com",
         "b.example.com",
         "c.example.com:1.2.3.4"
         "d.example.com:1.2.3.4,10.0.0.1"
        ]
        
    If a line includes a : delimiter, strip off the ips and add them to the database
    '''
    def add_domains(self, domains):
        (copy, inscope, outscope) = self.api.get_program_scope(self.get_program())
        add_domains = {}
        add_inscope = []
        
        # Keep a copy of the blacklist for reference
        blacklist = self.get_blacklist(copy=copy)
        
        for domain in domains:
            blacklisted_ip = False
            ips = []
            domain = domain.lower()
            
            if ':' in domain:
                domain, ips = domain.split(':',1)
                ips = ips.split(',')
                
                for ip in ips:
                    if ip in blacklist:
                        blacklisted_ip = True
                    if not REGEX_IP.match(ip):
                        ips.remove(ip)
                    
            
            # housekeeping
            if domain.endswith('.'):
                domain = domain[:-1]
            
            # not entirely sure this will ever occur, but hey (update: it does occur, as a result of crt.sh)
            # it makes sense to do this here, because it will still check whether it is in scope
            # before extending the existing scope.
            if domain.startswith('*.'):
                domain = domain[2:]
                # if it matches the existing scope definition,
                # add this wildcard to the scope too
                if REGEX_DOMAIN.match(domain) and not self.matches_scope(domain, outscope) and self.matches_scope(domain, inscope):
                    add_inscope.append('*.'+domain)
                    
            # Avoid adding blacklisted domains or
            # domains that resolve to blacklisted ips
            if blacklisted_ip or domain in blacklist:
                self.debug('blacklisted: '+domain)
                continue
            # It must match the format of a domain name
            if not REGEX_DOMAIN.match(domain):
                self.debug('REGEX_DOMAIN failed: '+domain)
                continue
            # It may not be explicitly outscoped unless --ignore-scope is set
            if not self.arguments['--ignore-scope'] and self.matches_scope(domain, outscope):
                self.debug('outscope: '+domain)
                continue
            # It must match the in scope, except if we're trying to @INFER the program later,
            # which means we cannot verify the scope here
            # ur unless --ignore-scope is set
            if not self.get_program() == '@INFER' and not self.arguments['--ignore-scope'] and not self.matches_scope(domain, inscope):
                self.debug('Not inscope: '+domain)
                continue
                
            # leading underscores are not allowed in couchdb,
            # so we need to somehow escape them.
            if domain.startswith('_'):
                domain = '.'+domain
                
            # Add the ips if we already parsed other ips for this domain,
            # or otherwise create a new one
            if domain in add_domains and type(add_domains[domain]) is list:
                add_domains[domain].extend(ips)
            else:
                add_domains[domain] = ips
                
        # add all new scope at once to drastically reduce runtime of large input
        if len(add_inscope) > 0:
            self.add_inscope(add_inscope, passalong={'doc':copy, 'inscope':inscope, 'outscope':outscope})
        
        if len(add_domains) > 0:
            success, _ = self.api.add_documents('domain', add_domains, self.get_program(), source=self.arguments['-s'], tags=self.arguments['-t'])
        
            if self.arguments['--show-new'] and success:
                return ["[NEW] "+(x if not x.startswith('._') else x[1:]) for x in success if x]
    
    '''
    Use the bulk update function to mass remove domains.
    '''
    def remove_domains(self, domains):
        
        remove = {(domain if not domain.startswith('_') else '.'+domain ): {'_deleted': True} for domain in domains}
        removed = self.api.update_documents('domain', remove)
        
        if self.arguments['--show-new'] and removed:
            return ["[DELETED] "+(x if not x.startswith('._') else x[1:]) for x in removed if x]
        
    '''
    Remove all documents related to a program.
    '''
    def remove_program(self, program):
        all_docs = self.api.get_all_program_documents(program)
        remove_domains = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'domain' }
        remove_ips = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'ip' }
        remove_urls = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'url' }
        remove_services = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'service' }
        
        self.api.update_documents('domain', remove_domains)
        self.api.update_documents('ip', remove_ips)
        self.api.update_documents('url', remove_urls)
        self.api.update_documents('service', remove_services)
        self.api.update_documents('program', {program: {'_deleted': True}})

    '''
    Update properties of a domain
    '''
    def update_domains(self, domains):
        update_domains = {}
            
        for domain in domains:
            
            ips = []
            domain = domain.lower()
            
            # split ips if provided
            if ':' in domain:
                domain, ips = domain.split(':')
                ips = ips.split(',')
                
                for ip in ips:
                    if not REGEX_IP.match(ip):
                        ips.remove(ip)
                
            # housekeeping
            if domain.endswith('.'):
                domain = domain[:-1]
            # leading underscores are not allowed in couchdb,
            # so we need to somehow escape them.
            if domain.startswith('_'):
                domain = '.'+domain
                
            # Add the ips if we already parsed other ips for this domain,
            # or otherwise create a new one
            if domain in update_domains and type(update_domains[domain]['ips']) is list:
                update_domains[domain]['ips'].extend(ips)
            else:
                update_domains[domain] = {"ips": ips}
            
            if(self.arguments['-t'] and len(self.arguments['-t']) > 0):
                update_domains[domain]['tags'] = self.api.process_tags(self.arguments['-t'])
                
            if self.arguments['-s']:
                update_domains[domain]['source'] = self.arguments['-s']
        
        updated = self.api.update_documents('domain', {domain: update_domains[domain] for domain in update_domains.keys()}, append_tags=self.arguments['--append-tags'])
        
        if self.arguments['--show-new'] and updated:
            return [ "[UPDATED] "+(x if not x.startswith('._') else x[1:]) for x in updated if x]
    
    def add_ips(self, ips):
        add_ips = {}
        
        # Keep a copy of the blacklist for reference
        blacklist = self.get_blacklist()
        
        for ip in ips:
            
            blacklisted_domain = False
            domains = []
            
            if ':' in ip:
                ip, domains = ip.split(':')
                domains = domains.split(',')
                
                for domain in domains:
                    if not REGEX_DOMAIN.match(domain):
                        domains.remove(domain)
#                    if domain in blacklist:
#                        blacklisted_domain = True
            
            if blacklisted_domain:
                continue
            if ip in blacklist:
                continue
            if not REGEX_IP.match(ip):
                continue
           
            if ip in add_ips and type(add_ips[ip]) is list:
                add_ips[ip].extend(domains)
            else: 
                add_ips[ip] = domains

        success, _ = self.api.add_documents('ip', add_ips, self.get_program(), source=self.arguments['-s'], tags=self.arguments['-t'])
        
        if self.arguments['--show-new'] and success:
            return ["[NEW] "+x for x in success if x]
        
    def remove_ips(self, ips):

        remove = {ip: {'_deleted': True} for ip in ips}
        removed = self.api.update_documents('ip', remove)
        
        if self.arguments['--show-new'] and removed:
            return ["[DELETED] "+x for x in removed if x] 
            
    '''
    Update properties of an IP
    '''
    def update_ips(self, ips):
        
        update_ips = {}
            
        for ip in ips:
            
            domains = []
            if ':' in ip:
                ip, domains = ip.split(':')
                domains = domains.split(',')
                
            # housekeeping
            updated_domains = []
            for domain in domains:
                if domain.endswith('.'):
                    domain = domain[:-1]
                if REGEX_DOMAIN.match(domain):
                    updated_domains.append(domain)
            
            # Add the domains if we already parsed other domains for this ip,
            # or otherwise create a new one
            if ip in update_ips and type(update_ips[ip]['domains']) is list:
                update_ips[ip]['domains'].extend(updated_domains)
            else:
                update_ips[ip] = {"domains": updated_domains}
            
            if(self.arguments['-t'] and len(self.arguments['-t']) > 0):
                update_ips[ip]['tags'] = self.api.process_tags(self.arguments['-t'])
                
            if self.arguments['-s']:
                update_ips[ip]['source'] = self.arguments['-s']
            
        updated = self.api.update_documents('ip', {ip: update_ips[ip] for ip in update_ips.keys()}, append_tags=self.arguments['--append-tags'])
        
        if self.arguments['--show-new'] and updated:
            return [ "[UPDATED] "+x for x in updated if x]
    
    '''
    The RUBYX client is responsible for ensuring the added urls are added to the appropriate hostname (domain or IP),
    unless a hostname is explicitly provided.
    
    If spaces are found, the input will be split in three parts: path, status code and response size, and the tool will 
    store those as part of the url.
    
    Expected format:
        urls = [
         "http://a.example.com:8080/test",
         "//b.example.com/example",
         "/robots.txt",
         "/page 200 9209",
         "http://hostname.com/page 401 170"
        ]
            
    The database stores urls as individual documents with:
     - identifier (url)
     - status
     - content_length
     - hostname (referencing a domain)
     - program
     
    '''
    def add_urls(self, urls):
        
        (_, inscope, outscope) = self.api.get_program_scope(self.get_program())
        add_urls = {}
        
        for url in urls:
            parts = url.split(' ')
            url = parts[0]
            
            # To support urls without a schema that aren't relative URLS
            # e.g. '    www.example.com', add a leading protocol 
            if not url.startswith('http://') and not url.startswith('https://') and not url.startswith('/'):
                url = 'http://'+url
                
            # parse properties of the URL
            u = urlparse(url)
            scheme = u.scheme
            port = u.port
            query = u.query if u.query else None
            
            # Usually we parse the hostname from the URL,
            # but we don't need to if a hostname is explicitly set
            if not self.arguments['-d']:
                hostname = u.hostname
            else:
                hostname = self.arguments['-d']
                                          
            # It's still possible hostname is empty, e.g. for
            # rubyx url add '/relative' without a -d hostname set;
            # We can't process relative URLS without understanding,
            # so need to skip those for now. Better ideas wecome!
            if not hostname:
                self.debug("Hostname could not be parsed, skipping "+url)
                continue
            
            # If the provided hostname in -d does not match the parsed hostname,
            # we won't add it to avoid polluting the dataset
            if u.hostname and not u.hostname == hostname:
                self.debug("Provided hostname "+hostname+" did not match parsed hostname "+u.hostname+", skipping...")
                continue
                
            # If the provided URL is relative, we need to rewrite the URL
            # with the provided hostname, and we will ALWAYS assume http port 80
            if not u.netloc:
                url = 'http://' + hostname + url
                port = 80
                scheme = 'http'
            if url.startswith('//'):
                url = 'http:' + url
                port = 80
                scheme = 'https'
            if '?' in url:
                url = url.split('?')[0]
                
            # It must match the format of a domain name or an IP address
            if not REGEX_DOMAIN.match(hostname) and not REGEX_IP.match(hostname):
                self.debug("Illegal hostname: "+hostname)
                continue
            # It may not be explicitly outscoped, unless --ignore-scope is set
            if not self.get_program() == '@INFER' and not self.arguments['--ignore-scope'] and self.matches_scope(hostname, outscope):
                self.debug("skipping outscoped hostname: "+hostname)
                continue
            # It must match the in scope, unless --ignore-scope is set
            if not self.get_program() == '@INFER' and not self.arguments['--ignore-scope'] and not self.matches_scope(hostname, inscope):
                self.debug("skipping not inscope hostname: "+hostname)
                continue
            
            if not port:
                if scheme == 'http':
                    port = 80
                if scheme == 'https':
                    port = 443
            
            # remove explicit ports 80 and 443 to avoid duplicate values
            # when they are also added without the ports
            if port == 80 and ':80/' in url:
                url = url.replace(':80/', '/', 1)
            elif port == 443 and ':443/' in url:
                url = url.replace(':443/', '/', 1)
            
            if url in add_urls and query and query not in add_urls[url]['query']:
                add_urls[url]['query'].append(query)
            else:
                if len(parts) == 1: # only a url
                    if not url in add_urls:
                        add_urls[url] = {
                            "hostname": hostname,
                            "port": int(port),
                            "query": [query] if query else [],
                            "scheme": scheme,
                            "path": u.path
                        }

                elif len(parts) == 3: # url, status code and content length
                    add_urls[url] = {
                        "hostname": hostname,
                        "port": int(port),
                        "status": int(parts[1]),
                        "content_length": int(parts[2]),
                        "query": [query] if query else [],
                        "scheme": scheme,
                        "path": u.path
                    }
        
        success, failed = self.api.add_documents('url', add_urls, self.get_program(), source=self.arguments['-s'], tags=self.arguments['-t'])
        
        # assuming the failed updates were the result of duplicates, try bulk updating the url that failed
        # but first add the tags to the document in case they need to be updated too
        if(self.arguments['-t'] and len(self.arguments['-t']) > 0):
            for url in failed:
                add_urls[url]['tags'] = self.api.process_tags(self.arguments['-t'])
        updated = self.api.update_documents('url', {url: add_urls[url] for url in failed}, append_tags=self.arguments['--append-tags'])
        
        
        results = []
        if self.arguments['--show-new'] and updated:
            results = [ "[UPDATED] "+x for x in updated if x]
        if self.arguments['--show-new'] and success:
            results = results + ["[NEW] "+x for x in success if x]
        if len(results) > 0:
            return results
        
    '''
    Remove URLs
    '''
    def remove_urls(self, urls):
        
        # parse first comma-seperated value as the url to remove
        # to support `rubyx urls -d example.com | rubyx url remove -`
        
        remove = {url.split(" ")[0]: {'_deleted': True} for url in urls}
        _ = self.api.update_documents('url', remove)
        
    '''
    Remove Services
    '''
    def remove_services(self, services):
        remove = {service: {'_deleted': True} for service in services}
        _ = self.api.update_documents('service', remove)
        
        
    '''
    Store services (i.e. open ports) of IP addresses
    
    The identifier will always 
    
    Expected format:
    
        services = [
         "127.0.0.1:80",
         "127.0.0.1:80:http",
         "127.0.0.1:443,
         "127.0.0.1:8443:https"
        ]
    '''
    def add_services(self, services):
        
        # TODO: should this be matched against scope?
        # RUBYX doesn't currently support IPs or CIDR as scope, so
        
        # this would require some additional features a that level.
        # 
        # (inscope, outscope) = self.api.get_program_scope(self.get_program())
        add_services = {}
        for service in services:
            
            ip = None
            port = None
            sname = None
            
            if ':' in service:
                ip, port = service.split(':', 1)
                
                if ':' in port:
                    port, sname = port.split(':', 1)

            if not ip or not port:
                continue
            
            if not REGEX_IP.match(ip):
                continue
            if not port.isnumeric():
                continue
            if int(port) < 0 or int(port) > 65535:
                continue
            
            sid = ip+':'+port
           
            add_services[sid] = {'ip': ip, 'port': port}
            if self.arguments['-s']:
                add_services[sid]['source'] = self.arguments['-s']
                
            if sname:
                add_services[sid]['service'] = sname

        success, failed = self.api.add_documents('service', add_services, self.get_program(), source=self.arguments['-s'], tags=self.arguments['-t'])
        
        # assuming the failed updates were the result of duplicates, try bulk updating the services that failed
        # but first add the tags to the document in case they need to be updated too
        if(self.arguments['-t'] and len(self.arguments['-t']) > 0):
            for sid in failed:
                add_services[sid]['tags'] = self.api.process_tags(self.arguments['-t'])
                
        updated = self.api.update_documents('service', {sid: add_services[sid] for sid in failed}, append_tags=self.arguments['--append-tags'])
        
        results = []
        if self.arguments['--show-new'] and updated:
            results = [ "[UPDATED] "+x for x in updated if x]
        if self.arguments['--show-new'] and success:
            results = results + ["[NEW] "+x for x in success if x]
        if len(results) > 0:
            return results
        
    def add_alerts(self, alerts):
        import time
        
        add_alerts = {}
        for alert in alerts:
            add_alerts[alert] = {
                'timestamp': time.time(),
                'message': alert # for backwards compatibility
            }
        
        success, failed = self.api.add_documents('alert', add_alerts, self.get_program(), source=self.arguments['-s'], tags=self.arguments['-t'])
        
        if self.arguments['--show-new'] and success:
            return ["[NEW] "+x for x in success if x]
        
    
    def disable_program(self, program):
        self.api.update_document("program", program, {"disabled":True})
        
    def enable_program(self, program):
        self.api.update_document("program", program, {"disabled":False})
        
    def add_inscope(self, elements, passalong={}):
        doc = passalong['doc'] if 'doc' in passalong else None  
        inscope = passalong['inscope'] if 'inscope' in passalong else None
        outscope = passalong['outscope'] if 'outscope' in passalong else None
        
        if not doc or not inscope or not outscope:
            (doc, inscope, outscope) = self.api.get_program_scope(self.get_program())
        changed = False
        
        for e in elements:
            e = e.lower()
            if e not in inscope:
                if REGEX_DOMAIN.match(e) or e.startswith('*.') and REGEX_DOMAIN.match(e[2:]):
                    changed = True
                    inscope.append(e)
                # support CIDR and IPs in scope
                elif REGEX_IP.match(e):
                    changed = True
                    inscope.append(e)
                # try to parse as a URL and consider the hostname/IP as inscope
                else:
                    u = urlparse(e).hostname
                    if u and u.lower() not in inscope and REGEX_DOMAIN.match(u.lower()):
                        changed = True
                        inscope.append(u.lower())
                    if u and u.lower() not in inscope and REGEX_IP.match(u.lower()):
                        changed = True
                        inscope.append(u.lower())
        
        if changed:
            self.api.update_program_scope(self.get_program(), inscope, outscope, program=doc)
    
    def remove_inscope(self, elements):
        (doc, inscope, outscope) = self.api.get_program_scope(self.get_program())
        
        changed = False
        
        for e in elements:
            e = e.lower()
            if e and e in inscope:
                changed = True
                inscope.remove(e)
                
        if changed:
            self.api.update_program_scope(self.get_program(), inscope, outscope, program=doc)
        
    def add_outscope(self, elements):
        (doc, inscope, outscope) = self.api.get_program_scope(self.get_program())
        
        changed = False
        
        for e in elements:
            e = e.lower()
            if e not in outscope:
                if REGEX_DOMAIN.match(e) or e.startswith('*.') and REGEX_DOMAIN.match(e[2:]):
                    changed = True
                    outscope.append(e)
                elif REGEX_IP.match(e):
                    changed = True
                    outscope.append(e)
                # try to parse as a URL and consider the hostname as outscope
                else:
                    u = urlparse(e).hostname.lower()
                    if u not in inscope and REGEX_DOMAIN.match(u):
                        changed = True
                        outscope.append(u)
                    elif u not in inscope and REGEX_IP.match(u):
                        changed = True
                        outscope.append(u)
        
        if changed:
            self.api.update_program_scope(self.get_program(), inscope, outscope, program=doc)
    
    def remove_outscope(self, elements):
        (doc, inscope, outscope) = self.api.get_program_scope(self.get_program())
        
        changed = False
        
        for e in elements:
            e = e.lower()
            if e and e in outscope:
                changed = True
                outscope.remove(e)
                
        if changed:
            self.api.update_program_scope(self.get_program(), inscope, outscope, program=doc)
    
    def list_ips(self, list_all = False):
        if list_all:
            return self.api.get_ips_by_program_name(show_disabled=self.arguments['--show-disabled'])
        return self.api.get_ips_by_program_name(self.get_program())
    
    def list_ips_no_cdn(self, list_all = False):
        if list_all:
            return self.api.get_ips_by_program_name(filter_cdn=True, show_disabled=self.arguments['--show-disabled'])
        return self.api.get_ips_by_program_name(program_name=self.get_program(), filter_cdn=True)
    
    def list_domains(self, list_all = False):
        if list_all:
            return self.api.get_domains_by_program_name(show_disabled=self.arguments['--show-disabled'])
        return [x if not x.startswith('._') else x[1:] for x in self.api.get_domains_by_program_name(self.get_program())]
    
    def list_urls(self, by, key = False):
        if by == "hostname":
            return self.api.get_urls_by_hostname(key, with_query=self.arguments['--with-query'], root_only=self.arguments['--root'])
        elif by == "program":
            return self.api.get_urls_by_program(key, with_query=self.arguments['--with-query'], root_only=self.arguments['--root'])
        elif by == "all":
            return self.api.get_urls_by_program(with_query=self.arguments['--with-query'], root_only=self.arguments['--root'], show_disabled=self.arguments['--show-disabled']) # An empty key will return all results
        else:
            return self.api.get_urls_by_program(self.get_program(), with_query=self.arguments['--with-query'], root_only=self.arguments['--root'])
        
    def list_services(self, list_all = False):
        if list_all:
            return self.api.get_services_by_program_name(show_disabled=self.arguments['--show-disabled'])
        return self.api.get_services_by_program_name(self.get_program())
    
    def list_documents_view(self, doctype, view, list_all = False):
        if list_all:
            return self.api.get_documents_view(None, doctype, view, show_disabled=self.arguments['--show-disabled'])
        return self.api.get_documents_view(self.get_program(), doctype, view)
    
    def get_blacklist(self, copy=None):
        return self.api.get_program_blacklist(self.get_program(), doc=copy)
    
    def add_blacklist(self, elements):
        blacklist = self.get_blacklist()
        
        for e in elements:
            if e not in blacklist:
                blacklist.append(e)
        
        self.api.update_program_blacklist(self.get_program(), blacklist)
    
    def remove_blacklist(self, elements):
        blacklist = self.get_blacklist()
        
        for e in elements:
            if e in blacklist:
                blacklist.remove(e)
                
        self.api.update_program_blacklist(self.get_program(), blacklist)
        
    def load_config(self):
        with open(os.path.expanduser(CONFIG_FILE)) as json_file:
            self.config = json.load(json_file)

    def save_config(self):
        if self.file_config:
            with open(os.path.expanduser(CONFIG_FILE), 'w') as outfile:
                json.dump(self.config, outfile)
            
    def listen_for_changes(self):
        self.api.listen_for_changes()
        
    def search_tags(self, doctype):
        # Always use the active program unless --all is specified
        program_name = self.get_program()
        if(self.arguments['--all']):
            program_name = False

        # get the results for each of the queries
        # and only keep the elements that match all of them
        results = []
        subresults = []
        # TODO: I'm pretty sure this doesn't work for multiple filters
        # because docopt cannot distinguish between the subqueries,
        # it will always assume 'before' in e.g. rubyx programs where x is before 1 and y is after 2
        # would be interpreted as rubyx programs where x is before 1 and y is BEFORE 2
        # not sure how to solve this atm...
        for i in range(len(self.arguments['<tag_name>'])):
            if(self.arguments['before']):
                subresults = self.api.search_tags_between(self.arguments['<tag_name>'][i], self.arguments['<value>'][i], 'before', doctype, program_name, show_disabled=self.arguments['--show-disabled'], show_empty_scope=self.arguments['--show-empty-scope'])
            elif(self.arguments['after']):
                subresults = self.api.search_tags_between(self.arguments['<tag_name>'][i], self.arguments['<value>'][i], 'after', doctype, program_name, show_disabled=self.arguments['--show-disabled'], show_empty_scope=self.arguments['--show-empty-scope'])
            else:
                subresults = [x if not x.startswith('._') else x[1:] for x in self.api.search_tags(self.arguments['<tag_name>'][i], self.arguments['<value>'][i], doctype, program_name, show_disabled=self.arguments['--show-disabled'], show_empty_scope=self.arguments['--show-empty-scope'])]
            # filter out anything that doesn't match previous subsearch
            if i == 0:
                results = subresults
            else:
                results = [s for s in subresults if s in results]
            # don't bother continuing if we're already out of results
            if len(results) == 0:
                break

        return results


    def list_tags(self, tagname):
        # Always use the active program unless --all is specified
        program_name = self.get_program()
        if(self.arguments['--all']):
            program_name = False
        return self.api.get_tags(tagname, program_name)
        
    '''
    Check whether a domain matches a scope
    '''
    @staticmethod
    def matches_scope(domain, scope):
        for s in scope:
            # literal match
            if s == domain:
                return True
            # x.example.com matches *.example.com
            if s.startswith('*.') and domain.endswith('.'+s[2:]):
                return True
            # CIDR match
            if REGEX_IP.match(domain) and REGEX_IP.match(s) and RubyxClient.ip_in_cidr(domain, s):
                return True
        return False    
    
    def debug(self, msg):
        if 'debug' in self.config and self.config['debug']:
            print('[DEBUG] '+msg)
        
    def run(self):
        
        import pprint
        pp = pprint.PrettyPrinter(indent=4)
        #pp.pprint(self.arguments)
        
        if not self.config:
            try:
                self.load_config()
            except Exception:
                print('[WARNING] Could not read config file - make sure it exists and is readable')

        if self.arguments['new']:
            self.new_program()

        if self.arguments['use']:
            self.use_program()
            
        if self.arguments['disable']:
            self.disable_program(self.arguments['<program>'][0])
            
        if self.arguments['enable']:
            self.enable_program(self.arguments['<program>'][0])
            
        if self.arguments['rm']:
            self.remove_program(self.arguments['<program>'][0])

        if self.arguments['programs']:
            if self.arguments['where']:
                return self.search_tags("program")
            return self.list_programs(self.arguments['--show-disabled'], self.arguments['--show-empty-scope'])
            
        if self.arguments['program']:
            if self.arguments['active']:
                return self.get_program()
            if self.arguments['update']:
                return self.update_programs(self.arguments['<program>'])
            else:
                return self.get_program()

        if self.arguments['domains']:
            if self.arguments['<view>']:
                return self.list_documents_view("domain", self.arguments['<view>'], self.arguments['--all'])
            elif self.arguments['--resolved']:
                if self.arguments['--no-private']:
                    return self.list_documents_view("domain", 'resolved_exclude_private', self.arguments['--all'])
                else:
                    return self.list_documents_view("domain", 'resolved', self.arguments['--all'])
            elif self.arguments['--unresolved']:
                return self.list_documents_view("domain", 'unresolved', self.arguments['--all'])
            elif self.arguments['where']:
                return self.search_tags("domain")
            else:
                return self.list_domains(self.arguments['--all'])

        if self.arguments['domain']:
            if self.arguments['add']:
                if self.arguments['<domain>']:
                    return self.add_domains(self.arguments['<domain>'])
                elif self.arguments['-']:
                    return self.add_domains(process_stdin())
            if self.arguments['remove']:
                if self.arguments['<domain>']:
                    return self.remove_domains(self.arguments['<domain>'])
                elif self.arguments['-']:
                    return self.remove_domains(process_stdin())
            if self.arguments['update']:
                if self.arguments['<domain>']:
                    return self.update_domains(self.arguments['<domain>'])
                elif self.arguments['-']:
                    return self.update_domains(process_stdin())

        if self.arguments['ips']:
            if self.arguments['--filter-cdns']:
                return self.list_ips_no_cdn(self.arguments['--all'])
            elif self.arguments['where']:
                return self.search_tags("ip")
            return self.list_ips(self.arguments['--all'])

        if self.arguments['ip']:
            if self.arguments['add']:
                if self.arguments['<ip>']:
                    return self.add_ips(self.arguments['<ip>'])
                elif self.arguments['-']:
                    return self.add_ips(process_stdin())
            if self.arguments['remove']:
                if self.arguments['<ip>']:
                    return self.remove_ips(self.arguments['<ip>'])
                elif self.arguments['-']:
                    return self.remove_ips(process_stdin())
            if self.arguments['update']:
                if self.arguments['<ip>']:
                    return self.update_ips(self.arguments['<ip>'])
                elif self.arguments['-']:
                    return self.update_ips(process_stdin())

        if self.arguments['inscope']:
            if self.arguments['add']:
                if self.arguments['<element>']:
                    return self.add_inscope(self.arguments['<element>'])
                elif self.arguments['-']:
                    return self.add_inscope(process_stdin())
            if self.arguments['remove']:
                if self.arguments['<element>']:
                    return self.remove_inscope(self.arguments['<element>'])
                elif self.arguments['-']:
                    return self.remove_inscope(process_stdin())
                    
        if self.arguments['outscope']:
            if self.arguments['add']:
                if self.arguments['<element>']:
                    return self.add_outscope(self.arguments['<element>'])
                elif self.arguments['-']:
                    return self.add_outscope(process_stdin())
            if self.arguments['remove']:
                if self.arguments['<element>']:
                    return self.remove_outscope(self.arguments['<element>'])
                elif self.arguments['-']:
                    return self.remove_outscope(process_stdin())
                    
        if self.arguments['url']:
            if self.arguments['add']:
                if self.arguments['<url>']:
                    return self.add_urls(self.arguments['<url>'])
                if self.arguments['-']:
                    return self.add_urls([u.rstrip() for u in process_stdin()])
            if self.arguments['remove']:
                if self.arguments['<url>']:
                    return self.remove_urls(self.arguments['<url>'])
                if self.arguments['-']:
                    return self.remove_urls([u.rstrip() for u in process_stdin()])
                    
        if self.arguments['urls']:
            if self.arguments['-d']:
                return self.list_urls("hostname",self.arguments['-d'])
            elif self.arguments['where']:
                return self.search_tags("url")
            elif self.arguments['<program>']:
                return self.list_urls("program",self.arguments['<program>'])
            elif self.arguments['--all']:
                return self.list_urls("all")
            else:
                return self.list_urls("self")
            
        if self.arguments['service']:
            if self.arguments['add']:
                if self.arguments['<service>']:
                    return self.add_services(self.arguments['<service>'])
                if self.arguments['-']:
                    return self.add_services([u.rstrip() for u in process_stdin()])
            if self.arguments['remove']:
                if self.arguments['<service>']:
                    return self.remove_services(self.arguments['<service>'])
                if self.arguments['-']:
                    return self.remove_services([s.rstrip() for s in process_stdin()])
                
            
        if self.arguments['services']:
            if self.arguments['where']:
                return self.search_tags("service")
            else:
                return self.list_services(self.arguments['--all'])
            
        if self.arguments['blacklist']:
            if self.arguments['add']:
                if self.arguments['<element>']:
                    return self.add_blacklist(self.arguments['<element>'])
                elif self.arguments['-']:
                    return self.add_blacklist(process_stdin())
            if self.arguments['remove']:
                if self.arguments['<element>']:
                    return self.remove_blacklist(self.arguments['<element>'])
                elif self.arguments['-']:
                    return self.remove_blacklist(process_stdin())
        
        if self.arguments['agents']:
            return self.list_agents()
            
        if self.arguments['agent']:
            if self.arguments['list']:
                return self.list_agents()
            if self.arguments['remove']:
                return self.remove_agents(self.arguments['<agent>'])
            if self.arguments['register']:
                return self.register_agents(self.arguments['<agent>'])
            if self.arguments['gateway']:
                if self.arguments['<url>']:
                    return self.api.set_agent_gateway(self.arguments['<url>'][0])
                else:
                    return json.loads(self.api.get_document('agents_api_gateway')).get('url')
        
        if self.arguments['run']:
            return self.api.run_agent(self.arguments['<agent>'][0], self.get_program())
        
        if self.arguments['show']:
            if not self.arguments['-']:
                # fall back to get_document if a single document is requested,
                # for backwards compatibility, so we continue to return a single object
                # rather than an array.       
                if(len(self.arguments['<document>']) == 1):
                    return self.api.get_document(self.arguments['<document>'][0])
                
                return self.api.get_documents(self.arguments['<document>'])
            else:
                # if list of documents comes from stdin, we presume the
                # user does not know it might only contain one identifier, and should
                # always handle the output as an array
                return self.api.get_documents(process_stdin())

        if self.arguments['remove']:
            if not self.arguments['-']:
                remove_ids = self.arguments['<document>']
            else:
                remove_ids = process_stdin()
            yn = 'N'
            if not self.arguments['--yes']:
                yn = input('WARNING! This will remove '+str(len(remove_ids))+' document(s) from your datastore and cannot be reverted.\nAre you sure you want to continue? [y/N] ')
            if yn.lower() == 'y' or self.arguments['--yes']:
                self.debug('Removing list of '+str(len(remove_ids))+' documents...')
                removed = self.api.remove_documents(remove_ids)
                self.debug('Successfully removed '+str(len(removed))+' documents')

        if self.arguments['listen']:
            self.listen_for_changes()
            
        if self.arguments['alert']:
            if self.arguments['<message>']:
                return self.add_alerts(self.arguments['<message>'])
            elif self.arguments['-']:
                return self.add_alerts(process_stdin())
                    
        if self.arguments['scope']:
            if self.arguments['filter']:
                return self.filter_scope()
            else:
                return self.get_scope()
            
        if self.arguments['tags']:
            return self.list_tags(self.arguments['<name>'])
        
        if self.arguments['server']:
            import getpass
            admin = input('Admin username: ')
            password = getpass.getpass('Password: ')
            self.api.server_upgrade(admin, password)

        try:
            self.save_config()
        except Exception:
            print('[WARNING] Could not write to config file - make sure it exists and is writable')
            
def process_stdin():
    return list(filter(lambda x: not re.match(r'^\s*$', x),  sys.stdin.read().split('\n')))
            
def main():
    try:
        arguments = docopt(__doc__, version=VERSION)
        result = RubyxClient(arguments).run()
        if result:
            if type(result) is list:
                print("\n".join(result))
            else:
                print(result)
    except Exception as e:
        print('[ERROR] '+str(e))
            
if __name__ == '__main__':
    main()